// Code is written by Jianing Wei
// Unauthorized use or copying is prohibited
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <time.h>

#include "../../src/list.h"
#include "../../src/allocate.h"
#include "../../src/MatrixCoding_Haar_2stage.h"

#define DELTA 0.02 // finest quantization interval
// for 2310x3088 image resolution it's 0.1
#define NUMOFSTAGES 2


void Usage(char *argv[])
{
  printf("Wrong number of arguments. Usage:\n");
  printf("%s parameter.txt [FineMatrixSaved.dat]\n", argv[0]);
  printf("parameter.txt is a text file specifying parameter values used in the program\n");
  printf("FineMatrixSaved.dat is an optional data file which saves the finest quantized matrix.\n");
  printf("Try again ...\n");
  exit(-1);
}


int main(int argc, char *argv[])
{
  int i, j, non0entry;
  int num_levels, height, width, block_size, block_num_levels;
  int height_padded, width_padded;
  float delta; 
  FILE *fp, *fq;
  struct parameter para;
  clock_t time_1, time_2, time_3, time_4; 
  double cputime;
  char scalefile[50];
  char matrixsavefile[50], bandinfo[50], finematrixfile[50];
  LIST **listmatrix, **coarse_listmatrix;
  int saved_matrix = 0;
  struct bandinfo band_data;

  if (argc<2 || argc>3) {
    Usage(argv); 
  }  
  if (argc==3) {
    saved_matrix = 1;
    strcpy(finematrixfile, argv[2]);
  }

  if ( (fp = fopen(argv[1], "r")) ) {
    fscanf(fp, "%f", &(para.a));
    fscanf(fp, "%f", &(para.b));
    fscanf(fp, "%f", &(para.f));
    fscanf(fp, "%f", &(para.g));
    fscanf(fp, "%f", &delta);
    fscanf(fp, "%d", &num_levels);
    fscanf(fp, "%d", &block_num_levels);
    fscanf(fp, "%d", &height);
    fscanf(fp, "%d", &width);
    fscanf(fp, "%d", &block_size);
    fscanf(fp, "%s", scalefile);
    fscanf(fp, "%s", matrixsavefile);
    fscanf(fp, "%s", bandinfo);
    fclose(fp);
  }
  else {
    printf("cannot open parameter file\n");
    return 1;
  }
  para.a = para.a*1e-6;
  para.b = para.b*1e-6;
  height_padded = ceil((double)height/block_size)*block_size;
  width_padded = ceil((double)width/block_size)*block_size;
  printf("delta = %f\n", delta);
  printf("Parameters: a=%f, b=%f, f=%f, g=%f\n", para.a,para.b,
	 para.f,para.g);
  printf("Number of levels for block transform %d\n", block_num_levels);
  printf("padded height = %d, padded width = %d\n", height_padded, width_padded);

  band_data.h_size = (int*) malloc(sizeof(int)*num_levels);
  band_data.v_size = (int*) malloc(sizeof(int)*num_levels);
  band_data.d_size = (int*) malloc(sizeof(int)*num_levels);
  band_data.scaling_factors = (float*) malloc(sizeof(float)*(1+3*num_levels));
  listmatrix = (LIST**) multialloc(sizeof(LIST), 2, height_padded, width_padded);
  coarse_listmatrix = (LIST**) multialloc(sizeof(LIST), 2, height_padded, width_padded);

  
  if ( (fp = fopen(scalefile, "r")) ) {
    for(i=0; i<1+3*num_levels; i++) {
      fscanf(fp, "%f", &(band_data.scaling_factors[i]));
      printf("scaling factor %d is %f\n", i, band_data.scaling_factors[i]);
    }
    fclose(fp);
  }
  else {
    printf("cannot open scaling factors file\n");
    exit(1);
  }


  if ( (fq = fopen(bandinfo, "r")) ) {
    for ( i=0; i<num_levels; i=i+1 ) {
      fscanf(fq, "%d", &(band_data.h_size[i]));
    }
    for ( i=0; i<num_levels; i=i+1 ) {
      fscanf(fq, "%d", &(band_data.v_size[i]));
    }
    for ( i=0; i<num_levels; i=i+1 ) {
      fscanf(fq, "%d", &(band_data.d_size[i]));
    }
    fscanf(fq, "%d", &(band_data.a_size));
    fclose(fq);
    printf("Done reading band size file %s\n", bandinfo);
  }
  else {
    printf("cannot open file %s to read\n", bandinfo);
    exit(1);
  }

  
  for ( i=0; i<height_padded; i++) {
    for ( j=0; j<width_padded; j++) {
      listmatrix[i][j].head = NULL;
      listmatrix[i][j].tail = NULL;
      listmatrix[i][j].num_pix = 0;
      coarse_listmatrix[i][j].head = NULL;
      coarse_listmatrix[i][j].tail = NULL;
      coarse_listmatrix[i][j].num_pix = 0;
    }
  }

  if (saved_matrix==0) {
    if (NUMOFSTAGES==2) {
      sprintf(finematrixfile, "fine_quantized_matrix_%dx%d.dat", height, width);
    }
    else {
      sprintf(finematrixfile, "onestage_fine_quantized_matrix_%dx%d.dat", height, width);
    }
    printf("Missing fine quantized matrix. Now creating one...\n");
    
    if (NUMOFSTAGES==2) {
      MatrixCoding( listmatrix, para, DELTA, 
		  num_levels, block_num_levels, height, width, 
		  block_size, band_data);
    }
    else {
      MatrixCoding_stage1( listmatrix, para, DELTA, 
		  num_levels, block_num_levels, height, width, 
		  block_size, band_data);
    }
  
  
    if ( (fq = fopen(finematrixfile, "w")) ) {
      WriteListmatrixHeader(fq, para, DELTA, num_levels, block_num_levels, 
	                    height_padded, width_padded, block_size);
      RecordListmatrix(fq, listmatrix, height_padded, width_padded);
      printf("successfully wrote fine quantized matrix into %s\n", finematrixfile);
      fclose(fq);
    }
    else {
      printf("cannot open %s to save matrix\n", finematrixfile);
      exit(1);
    }
  }
  else {
  
    if ( (fq=fopen(finematrixfile, "r")) ) {
      if (ReadNCheckListmatrixHeader(fq, para, DELTA, num_levels, block_num_levels, 
	                             height_padded, width_padded, block_size)) {
        ReadListmatrix(fq, listmatrix, height_padded, width_padded);
	// further quantize the matrix entries
        printf("successfully read listmatrix %s\n", finematrixfile);
      }
      else {
	printf("unable to read listmatrix %s\n", finematrixfile);
	exit(1);
      }
      fclose(fq);
    }
    else {
      printf("cannot open %s to read\n", finematrixfile);
      exit(1);
    }
  }

  FurtherQuantizeListmatrix( listmatrix, coarse_listmatrix, 
                             height_padded, width_padded, delta);

  if ( (fp = fopen(matrixsavefile, "w")) ) {
    WriteListmatrixHeader(fp, para, delta, num_levels, block_num_levels, 
	                    height_padded, width_padded, block_size);
    RecordListmatrix(fp, coarse_listmatrix, height_padded, width_padded);
    printf("successfully wrote final quantized matrix into %s\n", matrixsavefile);
    fclose(fp);
  } 
  else {
    printf("cannot open %s to write\n", matrixsavefile);
    exit(1);
  }


  non0entry = 0;
  for ( i=0; i<height_padded; i++) {
    for ( j=0; j<width_padded; j++) {
      non0entry = non0entry + coarse_listmatrix[i][j].num_pix;
    }
  }
  printf("non zero matrix entries in quantized matrix = %d\n", non0entry);

  
  
  for ( i=0; i<height_padded; i++) {
    for ( j=0; j<width_padded; j++) {
      ClearQueue(&listmatrix[i][j]);
      ClearQueue(&coarse_listmatrix[i][j]);
    }
  }
  multifree(listmatrix, 2);
  multifree(coarse_listmatrix, 2);
  free(band_data.h_size);
  free(band_data.v_size);
  free(band_data.d_size);
  free(band_data.scaling_factors);

  return 0;
}


