#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include "mex.h"
#include "../../../src/allocate.h"
#include "../../../src/SpaceVaryingConv.h"

mxArray* mxHaarTransform( const mxArray* mxImg, const mxArray* mxLevels )
{
  mxArray *mxOutput;
  double *pOutput;
  double *pImg = (double*) mxGetPr(mxImg);
  double *pLevels = (double*) mxGetPr(mxLevels);
  int levels = (int)(*pLevels);
  float **image, **wave_coef;
  int **band_size, **band_pos;
  const int *im_dims = mxGetDimensions(mxImg);
  int height, width, height_conved=0, width_conved=0;
  int i, j;
  int wt_dims[2];
  char filename[256];
  FILE *fp;

  if ( mxGetNumberOfDimensions(mxImg)!=2 ) {
    mexErrMsgTxt("Input image should be a single channel 2D image.\n");
  }
  if ( mxGetClassID(mxImg)!=mxDOUBLE_CLASS ) {
    mexErrMsgTxt("Input image should be of type double.");
  }

  height = im_dims[0];
  width = im_dims[1];
  band_size = multialloc(sizeof(int), 2, levels+1, 2);
  band_pos = multialloc(sizeof(int), 2, levels+1, 2);
  band_size[0][0] = height;
  band_size[0][1] = width;
  for ( i=1; i<=levels; i=i+1 ) {
    band_size[i][0] = (band_size[i-1][0]+1)/2;
    band_size[i][1] = (band_size[i-1][1]+1)/2;
  }
  band_pos[levels][0] = band_size[levels][0];
  band_pos[levels][1] = band_size[levels][1];
  for ( i=levels-1; i>=0; i=i-1 ) {
    band_pos[i][0] = band_pos[i+1][0]+band_size[i+1][0];
    band_pos[i][1] = band_pos[i+1][1]+band_size[i+1][1];
  }
  height_conved = band_pos[0][0];
  width_conved = band_pos[0][1];

  image = (float**) multialloc(sizeof(float), 2, height, width);
  wave_coef = (float**) multialloc(sizeof(float), 2, height_conved, width_conved);
  for (i=0; i<height; i++) {
    for (j=0; j<width; j++) {
      image[i][j] = (float)pImg[j*height+i];
    }
  }

  for (i=0; i<height_conved; i++) {
    for (j=0; j<width_conved; j++) {
      wave_coef[i][j] = 0;
    }
  }
  new_haar_transform(image, levels, 1, band_size, band_pos, wave_coef);
  wt_dims[0] = height_conved;
  wt_dims[1] = width_conved;
  multifree(band_size, 2);
  multifree(band_pos, 2);
  multifree(image, 2);
  mxOutput = mxCreateNumericArray(2, wt_dims, mxDOUBLE_CLASS, mxREAL);
  pOutput = (double*) mxGetPr(mxOutput);


  for (i=0; i<height_conved; i++) {
    for (j=0; j<width_conved; j++) {
      pOutput[j*height_conved+i] = (double) wave_coef[i][j];
    }
  }
  multifree(wave_coef, 2);
  return mxOutput;
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  if (nrhs!=2) {
    mexErrMsgTxt("Wrong number of inputs");
  }
  if (nlhs!=1) {
    mexErrMsgTxt("Wrong number of outputs");
  }

  plhs[0] = mxHaarTransform( prhs[0], prhs[1] );
}
