#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include "mex.h"
#include "../../../src/psf.h"

#define MAXVAL 2000000 /* max value used in power function */
#define SAMPLES 1048567 /* quantization levels in power LUT */
#define IMG_height 4.31

mxArray* StrayPsfImg(const mxArray* psf_para, const mxArray* img_height, 
                     const mxArray* img_width, const mxArray* position)
{
  mxArray *mxOutput;
  double *pOutput;
  double *para_array = (double*) mxGetPr(psf_para);
  double *p_height = (double*) mxGetPr(img_height);
  double *p_width = (double*) mxGetPr(img_width);
  double *p_position = (double*) mxGetPr(position);
  struct parameter para;
  int height, width;
  float x, y, x0, y0, x_org, y_org;
  struct power_LUT *powertable;
  float pixel_size;
  int i, j;
  int im_dims[2];

  para.a = para_array[0]*1e-6;
  para.b = para_array[1]*1e-6;
  para.f = para_array[2];
  para.g = para_array[3];
  height = (int)(*p_height);
  width = (int)(*p_width);
  im_dims[0] = height;
  im_dims[1] = width;
  x = p_position[0];
  y = p_position[1];
  powertable = build_power_LUT(para.f, MAXVAL, SAMPLES);
  pixel_size = IMG_height/(float)height;
  x_org = (float)(width-1)/2;
  y_org = (float)(height-1)/2;

  mxOutput = mxCreateNumericArray(2, im_dims, mxDOUBLE_CLASS, mxREAL);
  pOutput = (double*) mxGetPr(mxOutput);

  for (i=0; i<height; i++) {
    for (j=0; j<width; j++) {
      y0 = (float)i - y_org;
      x0 = (float)j - x_org;
      pOutput[j*height+i] = (double)computepsf(para, powertable, x0, y0, 
	  x, y, pixel_size);
    }
  }

  distroy_power_LUT(powertable);
  return mxOutput;
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  if (nrhs!=4) {
    mexErrMsgTxt("Wrong number of inputs");
  }
  if (nlhs!=1) {
    mexErrMsgTxt("Wrong number of outputs");
  }

  plhs[0] = StrayPsfImg( prhs[0], prhs[1], prhs[2], prhs[3] );
}
