clear all;
height = 256;
width = 256;
levels = 8;

% height = 1024;
% width = 1024;
% levels = 10;

% height = 2310;
% width = 3088;
% levels = 10;

num_points = 7;
total_points = num_points*num_points;

% % PSF parameter set 1
a = -165;
b = -835;
f = 1.21;
g = 0.015;

% % PSF parameter set 2
% a = 4000;
% b = 0;
% f = 1.15;
% g = 0.02;

% % real PSF parameters
% a = -16.5;
% b = -53.5;
% f = 1.31;
% g = 0.00176;

num_nonzero = 2000;
scaling_factor_file = 'scaling_factors_256x256_8levels.txt';
% scaling_factor_file = 'scaling_factors_1024x1024_10levels.txt';
% scaling_factor_file = 'scaling_factors_2310x3088_10levels.txt';

band_size(1,1) = height;
band_size(1,2) = width;
for l=1:levels
    band_size(l+1,1) = floor((band_size(l,1)+1)/2);
    band_size(l+1,2) = floor((band_size(l,2)+1)/2);
end

scaling = load(scaling_factor_file);
if (length(scaling)~=3*levels+1)
    error('dimension of scaling factor does not match number of levels\n');
end

h_size = zeros(total_points, levels);
v_size = zeros(total_points, levels);
d_size = zeros(total_points, levels);
a_size = zeros(total_points, 1);

x_org = (width+1)/2;
y_org = (height+1)/2;
gridx = linspace(0, width, num_points+2);
gridy = linspace(0, height, num_points+2);
n=0;
 
for i=1:num_points
    for j=1:num_points
        n = n + 1;
        x = gridx(i+1)-x_org;
        y = gridy(j+1)-y_org;
        psf = mxStrayPsf([a b f g], height, width, [x y]);
        psf_wt = mxHaarTransform(psf, levels);
        
        psf_wt_scaled = scale_matrix(psf_wt, scaling, band_size);
        sorted = sort(reshape(abs(psf_wt_scaled), size(psf_wt_scaled,1)*size(psf_wt_scaled,2), 1));
        threshold = sorted(length(sorted)-num_nonzero);
        bin_map = (abs(psf_wt_scaled)>=threshold);
        x_pos = 1+band_size(levels+1, 2);
        y_pos = 1+band_size(levels+1, 1);
        
        [u, v] = find(bin_map(1:band_size(levels+1,1), 1:band_size(levels+1,2)));
        mean_x = mean(v);
        mean_y = mean(u);
        max_radius = max((u-mean_y).^2+(v-mean_x).^2);
        a_size(n) = sqrt(max_radius);
        
        for l=1:levels
            [u, v] = find(bin_map(1:band_size(levels+2-l, 1), x_pos:x_pos+band_size(levels+2-l, 2)-1));
            if isempty(u)
                disp('no nonzero entries');
                h_size(n,levels-l+1) = 0;
            else
                mean_x = (1+band_size(levels+2-l, 2))/2 + x/2^(levels-l+1);
                mean_y = (1+band_size(levels+2-l, 1))/2 + y/2^(levels-l+1);
                max_radius = max((u-mean_y).^2+(v-mean_x).^2);
                h_size(n,levels-l+1) = sqrt(max_radius);
            end
            
            [u, v] = find(bin_map(y_pos:y_pos+band_size(levels+2-l, 1)-1, 1:band_size(levels+2-l, 2)));
            if isempty(u)
                disp('no nonzero entries');
                v_size(n,levels-l+1) = 0;
            else
                mean_x = (1+band_size(levels+2-l, 2))/2 + x/2^(levels-l+1);
                mean_y = (1+band_size(levels+2-l, 1))/2 + y/2^(levels-l+1);
                max_radius = max((u-mean_y).^2+(v-mean_x).^2);
                v_size(n,levels-l+1) = sqrt(max_radius);
            end
            
            [u, v] = find(bin_map(y_pos:y_pos+band_size(levels+2-l, 1)-1, x_pos:x_pos+band_size(levels+2-l, 2)-1));
            if isempty(u)
                disp('no nonzero entries');
                d_size(n,levels-l+1) = 0;
            else
                mean_x = (1+band_size(levels+2-l, 2))/2 + x/2^(levels-l+1);
                mean_y = (1+band_size(levels+2-l, 1))/2 + y/2^(levels-l+1);
                max_radius = max((u-mean_y).^2+(v-mean_x).^2);
                d_size(n,levels-l+1) = sqrt(max_radius);
            end
            
            x_pos = x_pos + band_size(levels+2-l, 2);
            y_pos = y_pos + band_size(levels+2-l, 1);
        end

    end
end

a_size = max(a_size);
h_size = max(h_size);
v_size = max(v_size);
d_size = max(d_size);
a_size = round(a_size+1);
h_size(levels) = h_size(levels)+1;
v_size(levels) = v_size(levels)+1;
d_size(levels) = d_size(levels)+1;
for i=1:levels-1
    if h_size(i)>2*h_size(i+1)
        h_size(i+1) = ceil(h_size(i)/2);
    end
    if v_size(i)>2*v_size(i+1)
        v_size(i+1) = ceil(v_size(i)/2);
    end
    if d_size(i)>2*d_size(i+1)
        d_size(i+1) = ceil(d_size(i)/2);
    end
end
h_size = ceil(h_size);
v_size = ceil(v_size);
d_size = ceil(d_size);
a_size = ceil(a_size);

filename = sprintf('wave_circle_size_f%4.2f_%dx%d_%d.txt', f, height, width, num_nonzero);
fid = fopen(filename, 'w');
fprintf(fid, '%d\n', h_size);
fprintf(fid, '%d\n', v_size);
fprintf(fid, '%d\n', d_size);
fprintf(fid, '%d\n', a_size);
fclose(fid);

