// Code is written by Jianing Wei
// Unauthorized use or copying is prohibited.
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include "mex.h"
#include "../../../src/allocate.h"
#include "../../../src/SpaceVaryingConv.h"
#include "../../../src/psf.h"

#define NUMEVALUATION 1000
#define MAXVAL 2000000
#define SAMPLES 1048567
#define IMG_HEIGHT 4.31




void GenerateRand2dIndex(unsigned int *index2d, int height, int width, int num_index)
{
  int i, j=RAND_MAX, n;
  for (n=0; j>0; j=(j>>1), n++);

  n = n/2<14 ? n/2 : 14;
  for (i=0; i<num_index; i++) {
    index2d[2*i] = (rand()>>n)*(height-1)/(RAND_MAX>>n);
    index2d[2*i+1] = (rand()>>n)*(width-1)/(RAND_MAX>>n);
  }  
}


static float ComputeNRMSE(float **approx_output, float *exact_output, 
                          unsigned int *pixel_index, int num_pixels)
{
  int i, j, n;
  float energy, sqdiff, nrmse;
  energy = 0;
  sqdiff = 0;
  for ( n=0; n<num_pixels; n=n+1 ) {
    i = pixel_index[2*n];
    j = pixel_index[2*n+1];
    energy += exact_output[n]*exact_output[n];
    sqdiff += (exact_output[n]-approx_output[i][j])*(exact_output[n]-approx_output[i][j]);
  }
  nrmse = sqrt(sqdiff/energy);
  return nrmse;
}

static void ConvertUINT8to2Darray(unsigned char *pimg, float **img2d, int height, int width)
{
	int i, j;
	unsigned char *puchar=pimg;

	for (j=0; j<width; j++) {
		for (i=0; i<height; i++) {
			img2d[i][j] = (float)(*(puchar++));
		}
	}
}

static void ConvertDOUBLEto2Darray(double *pimg, float **img2d, int height, int width)
{
	int i, j;
	double *pdouble = pimg;

	for (j=0; j<width; j++) {
		for (i=0; i<height; i++) {
			img2d[i][j] = (float)(*(pdouble++));
		}
	}
}

static void Convert2DarraytoDOUBLE(float **img2d, double *pimg, int height, int width)
{
	int i, j;
	double *pdouble = pimg;

	for (j=0; j<width; j++) {
		for (i=0; i<height; i++) {
			*(pdouble++) = (double)(img2d[i][j]);
		}
	}
}


mxArray* OnlineConv(const mxArray* input, const mxArray* listmatrixfile, const mxArray* scaling, 
                    mxArray *plhs[], int nlhs)
{
  mxArray *mxOutput;
  double *mxoutimg;
  const int *im_dims = mxGetDimensions(input);
  unsigned char *puchar_img;
  double *pdouble_img;
  double *scaling_factors_double = (double*) mxGetPr(scaling);
  const int *scaling_dims = mxGetDimensions(scaling);
  float *scaling_factors;
  int i, j;
  int height, width;
  struct parameter para;
  float delta;
  int num_levels, block_num_levels, block_size, height_padded, width_padded;
  float **input_img, **output_img;
  unsigned int pixel_index[NUMEVALUATION*2];
  float pixel_val[NUMEVALUATION*2];
  LIST **listmatrix;
  double* p_nrmse;
  char *pchar;
  char *listmatrix_file = mxArrayToString(listmatrixfile);
  FILE *fp;
  struct power_LUT *powertable;
  float pixel_size;

  // *mxNRMSE = mxCreateDoubleMatrix(1, 1, mxREAL);
  if (nlhs==2) {
    plhs[1] = mxCreateDoubleMatrix(1, 1, mxREAL); 
    p_nrmse = (double*) mxGetPr(plhs[1]);
  }
  height = im_dims[0];
  width = im_dims[1];
  if ( mxGetNumberOfDimensions(input)!=2 ) {
    mexErrMsgTxt("Input image should be a single channel 2D image.\n");
  }
  if ( mxGetClassID(listmatrixfile)!=mxCHAR_CLASS ) {
    mexErrMsgTxt("Input sparse matrix file must be a string.");
  }
  if ( (mxGetNumberOfDimensions(scaling)!=1) && (*(mxGetDimensions(scaling)+1)!=1)  ) {
    mexPrintf("scaling factor dims %d x %d\n", *(mxGetDimensions(scaling)), *(mxGetDimensions(scaling)+1));
    mexErrMsgTxt("Scaling factors must be in 1-D array.");
  }


  if ( fp=fopen(listmatrix_file, "r") ) {
    if ( ReadListmatrixHeader( fp, &para, &delta, &num_levels, 
                          &block_num_levels, &height_padded, &width_padded, &block_size ) ) {
      if ( scaling_dims[0]!=(1+3*num_levels) ) {
        mexErrMsgTxt("Scaling dimension mismatch.");
      }
      if ( height>height_padded || width>width_padded ) {
	mexErrMsgTxt("Input image dimension doesn't match stored matrix dimension.");
      }
      listmatrix = (LIST**) multialloc(sizeof(LIST), 2, height_padded, width_padded);
      for (i=0; i<height_padded; i++) {
	for (j=0; j<width_padded; j++) {
	  listmatrix[i][j].head = NULL;
	  listmatrix[i][j].tail = NULL;
	  listmatrix[i][j].num_pix = 0;
	}
      }
      ReadListmatrix( fp, listmatrix, height_padded, width_padded );
    }
    else {
      mexErrMsgTxt("Cannot read header of saved matrix\n");
    }
    fclose(fp);
  }
  else {
    printf("cannot open file %s to read\n", listmatrix_file);
    mexErrMsgTxt("Try again!\n");
  }
  
  
  input_img = multialloc(sizeof(float), 2, height, width);
  if ( mxGetClassID(input)==mxUINT8_CLASS ) {
	  puchar_img = (unsigned char *) mxGetPr(input);
	  ConvertUINT8to2Darray( puchar_img, input_img, height, width);
  }
  else {
    if ( mxGetClassID(input)==mxDOUBLE_CLASS ) {
  	    pdouble_img = (double *) mxGetPr(input);
	    ConvertDOUBLEto2Darray( pdouble_img, input_img, height, width );
    }
    else {
      mexErrMsgTxt("input image must be UINT8 or DOUBLE\n");
    }
  }
  
  scaling_factors = (float*) malloc(sizeof(float)*(1+3*num_levels));
  for (i=0; i<1+3*num_levels; i++) {
    scaling_factors[i] = (float)scaling_factors_double[i];
  }
  
  output_img = multialloc(sizeof(float), 2, height, width);
  SpaceVaryingTransform( input_img, height, width, listmatrix, 
                         height_padded, width_padded, block_size, block_num_levels,   
                         scaling_factors, num_levels, output_img, 2 );
 
  if (nlhs==2) { 
    srand(1);
    GenerateRand2dIndex(pixel_index, height, width, NUMEVALUATION);
    if (fp = fopen("imgnonzero_256x256_2.txt", "w")) {
      for (i=0; i<NUMEVALUATION; i++) {
        fprintf(fp, "%d %d\n", *(pixel_index+2*i), *(pixel_index+2*i+1));
      }
      fclose(fp);
    }
    else {
      printf("cannot open file to read\n");
      exit(-1);
    }

    powertable = build_power_LUT(para.f, MAXVAL, SAMPLES);
    pixel_size = IMG_HEIGHT/height;
    RigoConv( para, powertable, height, width, input_img, pixel_val, 
              pixel_index, NUMEVALUATION, pixel_size);
    distroy_power_LUT(powertable);
  
    *p_nrmse = ComputeNRMSE( output_img, pixel_val, pixel_index, NUMEVALUATION );
  }

  multifree(input_img, 2);
  multifree(listmatrix, 2);
  free(scaling_factors);

  mxOutput = mxCreateNumericArray(2, im_dims, mxDOUBLE_CLASS, mxREAL);
  mxoutimg = (double*) mxGetPr(mxOutput);
  Convert2DarraytoDOUBLE(output_img, mxoutimg, height, width);

  multifree(output_img, 2);
  return mxOutput;
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if (nrhs!=3) {
		mexErrMsgTxt("Wrong number of inputs\n");
	}
	if (nlhs>2) {
		mexErrMsgTxt("Wrong number of outputs\n");
	}

	plhs[0] = OnlineConv( prhs[0], prhs[1], prhs[2], plhs, nlhs );
}
