#include <stdlib.h>
#include "cimage.h"
#include "FastResize.h"
 
/* This function reduces image size by an integer factor */
void reduce_by_int(cimage_uchar* src, int factorx, int factory, cimage_uchar* dst)
{
    int w, h;
    int div = factorx * factory;
    int i, j, x, y;
	int val = 0;
    int rx, ry;
    unsigned char *line;
    unsigned char *dest_line;
 
    w = (src->width / factorx);
    rx = src->width - w*factorx;
    if (rx > 0) w++;
    h = (src->height / factory);
    ry = src->height - h*factory;
    if (ry > 0) h++;
  
	initUcharImage(dst, h, w);

	for (i=0; i<src->height; i+=factory) {
		dest_line = &(dst->ptr[(i/factory)*dst->width]);
		for (j=0; j<src->width; j+= factorx) {
			div = 0;
			val = 0;
			for (y=0; y<factory && i+y<src->height; y++) {
				line = &(src->ptr[(i+y)*src->width]);
				for (x=0; x<factorx && j+x<src->width; x++) {
					div++;
					val += (int)line[j+x];
				}
			}
			val /= div;
			dest_line[j/factorx] = (unsigned char)val;
		}
	}

}

/* This function reduces the image size when the size of src and dst are close (< factor 2) */
static void reduce_size(cimage_uchar* src, cimage_uchar* dst)
{
	int srcw = src->width;
    int srch = src->height;
    int dstw = dst->width;
    int dsth = dst->height;
    int src_dest_w = (int)(65536.0f*(float)(srcw-1)/(float)(dstw-1));
    int src_dest_h = (int)(65536.0f*(float)(srch-1)/(float)(dsth-1));
    int ci = 0;
    for (int i=0; i<dsth; i++)
    {
        int y1 = ci >> 16;
        int y2 = y1 < srch-1 ? y1+1 : srch-1;
        int yoffd = (ci >> 8) & 0xff;
        int yofft = 256-yoffd;
        const unsigned char * line1 = &(src->ptr[y1*srcw]);
        const unsigned char * line2 = &(src->ptr[y2*srcw]);
        unsigned char * dst_line = &(dst->ptr[i*dstw]);
 
        int cj = 0;
        for(int j=0; j<dstw; j++)
        {
            int x1 = cj >> 16;
            int x2 = x1<srcw-1 ? x1+1 : srcw-1;
 
            int c1 = line1[x1];
            int c2 = line1[x2];
            int c3 = line2[x1];
            int c4 = line2[x2];
 
            int xoffr = (cj >> 8) & 0xff;
			int xoffl = 256-xoffr;
            dst_line[j] = ((c1*xoffl+c2*xoffr)*yofft+(c3*xoffl+c4*xoffr)*yoffd) >> 16;
 
            cj+=src_dest_w;
        }
        ci+=src_dest_h;
    }
}

static void downsample(cimage_uchar* src, cimage_uchar* dst)
{
    float src_dest_w = (float)(src->width-1)/(float)(dst->width-1);
    float src_dest_h = (float)(src->height-1)/(float)(dst->height-1);
	cimage_uchar im_temp;

	if (src_dest_w>2 || src_dest_h>2 ) {
      reduce_by_int(src, (int)src_dest_w, (int)src_dest_h, &im_temp);
      reduce_size(&im_temp, dst);
	  destroyUcharImage(&im_temp);
	}
	else {
	  reduce_size(src, dst);
	}
 
}

static void upsample(cimage_uchar* src, cimage_uchar* dst)
{
	int srcw = src->width;
    int srch = src->height;
    int dstw = dst->width;
    int dsth = dst->height;
    int src_dest_w = (int)(65536.0f*(float)(srcw-1)/(float)(dstw-1));
    int src_dest_h = (int)(65536.0f*(float)(srch-1)/(float)(dsth-1));

    int ci = 0;
    for (int i=0; i<dsth; i++)
    {
        int y1 = ci >> 16;
        int y2 = y1 < srch-1 ? y1+1 : srch-1;
        int yoffd = (ci >> 8) & 0xff;
		int yoffu = 256 - yoffd;
		const unsigned char * line1 = &src->ptr[y1*srcw];
		const unsigned char * line2 = &src->ptr[y2*srcw];
        unsigned char * dest_line = &(dst->ptr[i*dstw]);
        int cj = 0;
        for(int j=0; j<dstw; j++)
        {
            int x1 = cj >> 16;
			int x2 = x1 < srcw-1 ?  x1+1 : srcw-1;
            int xoffr = (cj >> 8) & 0xff;
            int xoffl = 256 - xoffr;
            int c1 = line1[x1];
            int c2 = line1[x2];
            int c3 = line2[x1];
            int c4 = line2[x2];
            dest_line[j] = ((c1*xoffl+c2*xoffr)*yoffu+(c3*xoffl+c4*xoffr)*yoffd) >> 16;
            cj+=src_dest_w;
        }
        ci+=src_dest_h;
    }
 
}

void fast_resize(cimage_uchar* src, cimage_uchar* dst)
{
	if (src->height/dst->height > 1) {
		downsample(src, dst);
	}
	else {
		upsample(src, dst);
	}
}


/********* FLOATING POINT VERSION ***********/
void float_reduce_by_int(cimage* src, int factorx, int factory, cimage* dst)
{
    int w, h;
    int div = factorx * factory;
    int i, j, x, y;
    float val = 0.0;
    int rx, ry;
    float *line;
    float *dest_line;
 
    w = (src->width / factorx);
    rx = src->width - w*factorx;
    if (rx > 0) w++;
    h = (src->height / factory);
    ry = src->height - h*factory;
    if (ry > 0) h++;
  
    initImage(dst, h, w);

    for (i=0; i<src->height; i+=factory) {
	dest_line = &(dst->ptr[(i/factory)*dst->width]);
	for (j=0; j<src->width; j+= factorx) {
		div = 0;
		val = 0.0;
		for (y=0; y<factory && i+y<src->height; y++) {
			line = &(src->ptr[(i+y)*src->width]);
			for (x=0; x<factorx && j+x<src->width; x++) {
				div++;
				val += line[j+x];
			}
		}
		val /= (float)div;
		dest_line[j/factorx] = val;
	}
    }

}

/* This function reduces the image size when the size of src and dst are close (< factor 2) */
static void float_reduce_size(cimage* src, cimage* dst)
{
    int srcw = src->width;
    int srch = src->height;
    int dstw = dst->width;
    int dsth = dst->height;
    int src_dest_w = (int)(65536.0f*(float)(srcw-1)/(float)(dstw-1));
    int src_dest_h = (int)(65536.0f*(float)(srch-1)/(float)(dsth-1));
    int ci = 0;
    for (int i=0; i<dsth; i++)
    {
        int y1 = ci >> 16;
        int y2 = y1 < srch-1 ? y1+1 : srch-1;
        int yoffd = (ci >> 8) & 0xff;
        int yofft = 256-yoffd;
        const float * line1 = &(src->ptr[y1*srcw]);
        const float * line2 = &(src->ptr[y2*srcw]);
        float * dst_line = &(dst->ptr[i*dstw]);
 
        int cj = 0;
        for(int j=0; j<dstw; j++)
        {
            int x1 = cj >> 16;
            int x2 = x1<srcw-1 ? x1+1 : srcw-1;
 
            float c1 = line1[x1];
            float c2 = line1[x2];
            float c3 = line2[x1];
            float c4 = line2[x2];
 
            int xoffr = (cj >> 8) & 0xff;
            int xoffl = 256-xoffr;
            dst_line[j] = ((c1*(float)xoffl+c2*(float)xoffr)*yofft
		          +(c3*(float)xoffl+c4*(float)xoffr)*yoffd) / 65536.0;
 
            cj+=src_dest_w;
        }
        ci+=src_dest_h;
    }
}

static void float_downsample(cimage* src, cimage* dst)
{
    float src_dest_w = (float)(src->width-1)/(float)(dst->width-1);
    float src_dest_h = (float)(src->height-1)/(float)(dst->height-1);
    cimage im_temp;

    if (src_dest_w>2 || src_dest_h>2 ) {
      float_reduce_by_int(src, (int)src_dest_w, (int)src_dest_h, &im_temp);
      float_reduce_size(&im_temp, dst);
      destroyImage(&im_temp);
    }
    else {
      float_reduce_size(src, dst);
    }
 
}

static void float_upsample(cimage* src, cimage* dst)
{
    int srcw = src->width;
    int srch = src->height;
    int dstw = dst->width;
    int dsth = dst->height;
    int src_dest_w = (int)(65536.0f*(float)(srcw-1)/(float)(dstw-1));
    int src_dest_h = (int)(65536.0f*(float)(srch-1)/(float)(dsth-1));

    int ci = 0;
    for (int i=0; i<dsth; i++)
    {
        int y1 = ci >> 16;
        int y2 = y1 < srch-1 ? y1+1 : srch-1;
        int yoffd = (ci >> 8) & 0xff;
        int yoffu = 256 - yoffd;
	const float * line1 = &src->ptr[y1*srcw];
	const float * line2 = &src->ptr[y2*srcw];
        float * dest_line = &(dst->ptr[i*dstw]);
        int cj = 0;
        for(int j=0; j<dstw; j++)
        {
            int x1 = cj >> 16;
            int x2 = x1 < srcw-1 ?  x1+1 : srcw-1;
            int xoffr = (cj >> 8) & 0xff;
            int xoffl = 256 - xoffr;
            float c1 = line1[x1];
            float c2 = line1[x2];
            float c3 = line2[x1];
            float c4 = line2[x2];
            dest_line[j] = ((c1*(float)xoffl+c2*(float)xoffr)*yoffu
		           +(c3*(float)xoffl+c4*(float)xoffr)*yoffd) / 65536.0;
            cj+=src_dest_w;
        }
        ci+=src_dest_h;
    }
 
}

void fast_resize_float(cimage* src, cimage* dst)
{
	if (src->height/dst->height > 1) {
		float_downsample(src, dst);
	}
	else {
		float_upsample(src, dst);
	}
}
