/* Created by Jianing Wei */
/* ApplyScaling changed */
/* Dynamic Allocation of WAC changed  */
/* Extention mode needs to be zero extention to be orthonormal */
/* This code also changes the data structure for storing sparse column data */
/* A bug in FastComputeApproxCoef is fixed */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <time.h>

#include "sparseWT.h"
#include "power.h"
#include "list.h"
#include "sparselink.h"


#define MAXVAL 2000000 /* max value used in power function */
#define SAMPLES 1048567 /* quantization levels in power LUT */
#define FILTERLEN_L 2
#define FILTERLEN_H 2
#define IMG_HEIGHT 4.31
#define MATRIX_SAVED 0
#define GROUND_TRUTH 0

struct parameter {
  float a;
  float b;
  float f;
  float g;
};

typedef struct parameter PARAMETER;


float computepsf(PARAMETER para, struct power_LUT *powertable, 
		 float x0, float y0, float x, float y, float pixel_size)
{
  float r2, term1, term2, value, norm_constant;
  float temp[1];
  x = x*pixel_size;
  y = y*pixel_size;
  x0 = x0*pixel_size;
  y0 = y0*pixel_size;

  r2 = x0*x0+y0*y0;
 
  norm_constant = 3.1415926*(para.g+para.a*r2)*(para.g+para.b*r2)/(para.f-1);
  norm_constant = norm_constant/pixel_size/pixel_size;
 
  if (r2==0) {
    temp[0] = 1+(x*x+y*y)/(para.g*para.g);
    if (temp[0]>MAXVAL) {
      value = 0;
    }
    else {
      power(temp, 1, powertable);
      value = 1/temp[0]/norm_constant;
    }
  }
  else {
    term1 = (x*x0+y*y0-r2)/(para.g+para.a*r2);
    term2 = (-x*y0+y*x0)/(para.g+para.b*r2);
    temp[0] = 1+(term1*term1+term2*term2)/r2;
    if (temp[0]>MAXVAL) {
      value = 0;
    }
    else {
      power(temp, 1, powertable);
      value = 1/temp[0]/norm_constant;
    }
  }
  return value;

}

/* Rigorous convolution using traditional strategy */
void RigoConv(struct parameter para, struct power_LUT * powertable,  int height, int width, 
              float **img1, float **img2, int **imgentry, int imgnonzero, float pixel_size)
{
  int i, j, n;
  float x0, y0, x, y, x_org, y_org;
  float psf; 
  x_org = (float)(width-1)/2;
  y_org = (float)(height-1)/2;
  
  for ( i=0; i<height; i++) {
    for ( j=0; j<width; j++) {
      x0 = (float)j-x_org;
      y0 = (float)i-y_org;
      for ( n=0; n<imgnonzero; n=n+1 ) {
	x = (float)imgentry[n][1]-x_org;
	y = (float)imgentry[n][0]-y_org;
	psf = computepsf(para, powertable, x0, y0, x, y, pixel_size);
	img2[imgentry[n][0]][imgentry[n][1]] += psf*img1[i][j];
      }

    }
  }

}


int ZeroWaveletCoef(int k, int m, int n, float x, float y, int radius, 
		    float x_org, float y_org)
{
  x = x*pow(0.5, k)+x_org;
  y = y*pow(0.5, k)+y_org;
  if (x<0) {
    x = 0;
  }
  if (x>x_org*2) {
    x = x_org*2;
  }
  if (y<0) {
    y = 0;
  }
  if (y>y_org*2) {
    y = y_org*2;
  }
  if ((n-x)*(n-x)+(m-y)*(m-y)<=radius*radius) {
    return 1;
  }
  else {
    return 0;
  }

}

/* symmetric extention */
/*
float FastComputeApproxCoef(int k, int m, int n, float ***WAC, PARAMETER para, 
			    struct power_LUT *powertable, float x, float y, 
			    int height, int width, int *h_size, int *v_size, 
			    int *d_size, int **band_size, float pixel_size)
{
  float value = 0;
  float x_org, y_org, x_suborg, y_suborg;
  float a, h, v, d;
  int maxradius;

  if (k==0) {
    x_org = (float)(width-1)/2;
    y_org = (float)(height-1)/2;
    value = computepsf(para, powertable, (float)n-x_org, (float)m-y_org, x, y, pixel_size);
    WAC[k][m][n] = value;
    return value;
  }

  maxradius = h_size[k-1];
  if (maxradius<v_size[k-1]) {
    maxradius = v_size[k-1];
  }
  if (maxradius<d_size[k-1]) {
    maxradius = d_size[k-1];
  }
  x_suborg = (float)(band_size[k][1]-1)/2;
  y_suborg = (float)(band_size[k][0]-1)/2;
  if (ZeroWaveletCoef(k, m, n, x, y, maxradius, x_suborg, y_suborg)==0) {
    x_org = (float)(width-1)/2;
    y_org = (float)(height-1)/2;
    value = computepsf(para, powertable, (float)(n<<k)-x_org, (float)(m<<k)-y_org, x, y, pixel_size);
    value = value*pow(2,k);
    WAC[k][m][n] = value;
    return value;
  }
  else {
    a = FastComputeApproxCoef(k-1, 2*m, 2*n, WAC, para, powertable, 
			      x, y, height, width, h_size, v_size, d_size, 
			      band_size, pixel_size);
    if ( 2*n+1>=band_size[k-1][1] ) {
      h = a;
    }
    else {
      h = FastComputeApproxCoef(k-1, 2*m, 2*n+1, WAC, para, powertable, 
				x, y, height, width, h_size, v_size, d_size, 
				band_size, pixel_size);
    }
    if ( 2*m+1>=band_size[k-1][0] ) {
      v = a;
    }
    else {
      v = FastComputeApproxCoef(k-1, 2*m+1, 2*n, WAC, para, powertable, 
				x, y, height, width, h_size, v_size, d_size, 
				band_size, pixel_size);
    }
    if ( 2*n+1>=band_size[k-1][1] ) {
      d = v;
    }
    else {
      if ( 2*m+1>=band_size[k-1][0] ) {
	d = h;
      }
      else {
	d = FastComputeApproxCoef(k-1, 2*m+1, 2*n+1, WAC, para, powertable, 
				  x, y, height, width, h_size, v_size, d_size, 
				  band_size, pixel_size);
      }
    }

    value = (a+h+v+d)/2;
    WAC[k][m][n] = value;
    return value;
  }

}
*/

/* zero extention */
float FastComputeApproxCoef(int k, int m, int n, float ***WAC, PARAMETER para, 
			    struct power_LUT *powertable, float x, float y, 
			    int height, int width, int *h_size, int *v_size, 
			    int *d_size, int **band_size, float pixel_size)
{
  float value = 0;
  float x_org, y_org, x_suborg, y_suborg;
  float a, h, v, d;
  int maxradius;

  if (k==0) {
    x_org = (float)(width-1)/2;
    y_org = (float)(height-1)/2;
    value = computepsf(para, powertable, (float)n-x_org, (float)m-y_org, x, y, pixel_size);
    WAC[k][m][n] = value;
    return value;
  }

  maxradius = h_size[k-1];
  if (maxradius<v_size[k-1]) {
    maxradius = v_size[k-1];
  }
  if (maxradius<d_size[k-1]) {
    maxradius = d_size[k-1];
  }
  x_suborg = (float)(band_size[k][1]-1)/2;
  y_suborg = (float)(band_size[k][0]-1)/2;
  if (ZeroWaveletCoef(k, m, n, x, y, maxradius, x_suborg, y_suborg)==0) {
    x_org = (float)(width-1)/2;
    y_org = (float)(height-1)/2;
    value = computepsf(para, powertable, (float)(n<<k)-x_org, (float)(m<<k)-y_org, x, y, pixel_size);
    value = value*pow(2,k);
    WAC[k][m][n] = value;
    return value;
  }
  else {
    a = FastComputeApproxCoef(k-1, 2*m, 2*n, WAC, para, powertable, 
			      x, y, height, width, h_size, v_size, d_size, 
			      band_size, pixel_size);
    if ( 2*n+1>=band_size[k-1][1] ) {
      h = 0;
      d = 0;
    }
    else {
      h = FastComputeApproxCoef(k-1, 2*m, 2*n+1, WAC, para, powertable, 
				x, y, height, width, h_size, v_size, d_size, 
				band_size, pixel_size);
    }
    if ( 2*m+1>=band_size[k-1][0] ) {
      v = 0;
      d = 0;
    }
    else {
      v = FastComputeApproxCoef(k-1, 2*m+1, 2*n, WAC, para, powertable, 
				x, y, height, width, h_size, v_size, d_size, 
				band_size, pixel_size);
    }
    if ( (2*n+1<band_size[k-1][1]) && (2*m+1<band_size[k-1][0]) ) {
	d = FastComputeApproxCoef(k-1, 2*m+1, 2*n+1, WAC, para, powertable, 
				  x, y, height, width, h_size, v_size, d_size, 
				  band_size, pixel_size);
    }

    value = (a+h+v+d)/2;
    WAC[k][m][n] = value;
    return value;
  }

}




/* symmetric extention */
/*
void ExtWAC(float ***WAC, int k, int i, int j, int **band_size)
{
  float a;
  a = WAC[k-1][2*i][2*j];
  if ( 2*j+1>=band_size[k-1][1] ) {
    WAC[k-1][2*i][2*j+1] = a;
  }
  if ( 2*i+1>=band_size[k-1][0] ) {
    WAC[k-1][2*i+1][2*j] = a;
  }
  if ( 2*j+1>=band_size[k-1][1] ) {
    WAC[k-1][2*i+1][2*j+1] = WAC[k-1][2*i+1][2*j];
  }
  if ( 2*i+1>=band_size[k-1][0] ) {
    WAC[k-1][2*i+1][2*j+1] = WAC[k-1][2*i][2*j+1];
  }
}
*/

/* zero extention */
void ExtWAC(float ***WAC, int k, int i, int j, int **band_size)
{
  if ( 2*j+1>=band_size[k-1][1] ) {
    WAC[k-1][2*i][2*j+1] = 0;
    WAC[k-1][2*i+1][2*j+1] = 0;
  }
  if ( 2*i+1>=band_size[k-1][0] ) {
    WAC[k-1][2*i+1][2*j] = 0;
  }
}



void ComputeDetailCoef(float ***WAC, float **WaveCoef, int a_size, int *h_size, 
		       int *v_size, int *d_size, int levels, float x, 
		       float y, int **band_size, int **band_pos)
{
  int i, j, k;
  int subheight, subwidth, pos_row, pos_col, max_radius;
  float x_org, y_org, x0, y0;

  for ( k=1; k<=levels; k=k+1 ) {
    subheight = band_size[k][0];
    subwidth = band_size[k][1];
    pos_row = band_pos[k][0];
    pos_col = band_pos[k][1];
    x_org = (float)(subwidth-1)/2;
    y_org = (float)(subheight-1)/2;
    x0 = x_org + x*pow(0.5, k);
    y0 = y_org + y*pow(0.5, k);
    max_radius = h_size[k-1];
    if (v_size[k-1]>max_radius) {
      max_radius = v_size[k-1];
    }
    if (d_size[k-1]>max_radius) {
      max_radius = d_size[k-1];
    }
    for ( i=(int)y0-max_radius; i<=(int)y0+max_radius; i=i+1 ) {
      for ( j=(int)x0-max_radius; j<=(int)x0+max_radius; j=j+1 ) {
        if ( (i>=0) && (i<subheight) && (j>=0) && (j<subwidth) ) {
          ExtWAC(WAC, k, i, j, band_size);
        }
      }
    }

    for ( i=(int)y0-h_size[k-1]; i<=(int)y0+h_size[k-1]; i=i+1 ) {
      for ( j=(int)x0-h_size[k-1]; j<=(int)x0+h_size[k-1]; j=j+1 ) {
	if ( (ZeroWaveletCoef(k, i, j, x, y, h_size[k-1], x_org, y_org)==1) 
	     && i>=0 && i<subheight && j>=0 && j<subwidth ) {
	  WaveCoef[i][j+pos_col] = (WAC[k-1][2*i][2*j]-WAC[k-1][2*i][2*j+1]+
				     WAC[k-1][2*i+1][2*j]-WAC[k-1][2*i+1][2*j+1])/2;
	}
      }
    }
    for ( i=(int)y0-v_size[k-1]; i<=(int)y0+v_size[k-1]; i=i+1 ) {
      for ( j=(int)x0-v_size[k-1]; j<=(int)x0+v_size[k-1]; j=j+1 ) {
	if ( (ZeroWaveletCoef(k, i, j, x, y, v_size[k-1], x_org, y_org)==1) 
	     && i>=0 && i<subheight && j>=0 && j<subwidth) {
	  WaveCoef[i+pos_row][j] = (WAC[k-1][2*i][2*j]+WAC[k-1][2*i][2*j+1]-
				      WAC[k-1][2*i+1][2*j]-WAC[k-1][2*i+1][2*j+1])/2;
	}
      }
    }
    for ( i=(int)y0-d_size[k-1]; i<=(int)y0+d_size[k-1]; i=i+1 ) {
      for ( j=(int)x0-d_size[k-1]; j<=(int)x0+d_size[k-1]; j=j+1 ) {
	if ( (ZeroWaveletCoef(k, i, j, x, y, d_size[k-1], x_org, y_org)==1) 
	     && i>=0 && i<subheight && j>=0 && j<subwidth) {
	  WaveCoef[i+pos_row][j+pos_col] = (WAC[k-1][2*i][2*j]-WAC[k-1][2*i][2*j+1] -
					       WAC[k-1][2*i+1][2*j]+WAC[k-1][2*i+1][2*j+1])/2;
	}
      }
    }
  }

  for ( i=0; i<band_size[levels][0]; i=i+1 ) {
    for ( j=0; j<band_size[levels][1]; j=j+1 ) {
      if ( ZeroWaveletCoef(levels, i, j, x, y, a_size, x_org, y_org) ) {
	WaveCoef[i][j] = WAC[levels][i][j];
      }
    }
  }

}


void createflag(int k, int m, int n, int *h_size, int *v_size, 
		 int *d_size, int height, int width, float x, float y, 
		int ***flag_matrix)
{
  int maxradius;
  float x_suborg, y_suborg;
  if (k==0) {
    flag_matrix[k][m][n] = 1;
    return;
  }

  maxradius = h_size[k-1];
  if (maxradius<v_size[k-1]) {
    maxradius = v_size[k-1];
  }
  if (maxradius<d_size[k-1]) {
    maxradius = d_size[k-1];
  }
  x_suborg = (float)((width>>k)-1)/2;
  y_suborg = (float)((height>>k)-1)/2;
  if (ZeroWaveletCoef(k, m, n, x, y, maxradius, x_suborg, y_suborg)==0) {
    flag_matrix[k][m][n] = 1;
    return;
  }
  else {
    createflag(k-1,2*m,2*n,h_size,v_size,d_size,height,width,x,y,flag_matrix);
    createflag(k-1,2*m,2*n+1,h_size,v_size,d_size,height,width,x,y,flag_matrix);
    createflag(k-1,2*m+1,2*n,h_size,v_size,d_size,height,width,x,y,flag_matrix);
    createflag(k-1,2*m+1,2*n+1,h_size,v_size,d_size,height,width,x,y,flag_matrix);
    flag_matrix[k][m][n] = 1;
    return;
  }
}

int check_detail(int ***flag_matrix, int *h_size, int *v_size, 
		  int *d_size, int height, int width, float x, float y, int levels)
{
  int k, i, j;
  int maxradius;
  float x_suborg, y_suborg;
  float x0, y0;
  for ( k=1; k<=levels; k=k+1 ) {
    maxradius = h_size[k-1];
    if (maxradius<v_size[k-1]) {
      maxradius = v_size[k-1];
    }
    if (maxradius<d_size[k-1]) {
      maxradius = d_size[k-1];
    }
    x_suborg = (float)((width>>k)-1)/2;
    y_suborg = (float)((height>>k)-1)/2;
    x0 = x_suborg + x*pow(0.5, k);
    y0 = y_suborg + y*pow(0.5, k);
    for ( i=(int)y0-maxradius; i<=(int)y0+maxradius; i=i+1 ) {
      for ( j=(int)x0-maxradius; j<=(int)x0+maxradius; j=j+1 ) {
	if ( (ZeroWaveletCoef(k, i, j, x, y, maxradius, x_suborg, y_suborg)==1) && i>=0 && i<(height>>k) 
	    && j>=0 && j<(width>>k) ) {
	  if( flag_matrix[k-1][2*i][2*j]*flag_matrix[k-1][2*i][2*j+1]*flag_matrix[k-1][2*i+1][2*j]*flag_matrix[k-1][2*i+1][2*j+1] == 0) {
	    return 1;
	  }
	}

      }
    }

  }
  return 0;

}

/***** fast method for computing wavelet transform and store it in sparse form ******/
/*
void ComputeSparseWT(float ***WAC, LINKLIST ***Smatrix, int a_size, int *h_size, int *v_size, 
		     int *d_size, int levels, float x, float y, int row, int col, int block_size, 
                     float *scaling, int **band_size, int **band_pos)
{
  int i, j, k;
  int subheight, subwidth, blockrow, blockcol, pos_row, pos_col, max_radius;
  float x_org, y_org, x0, y0;
  float value;

  blockrow = row%block_size;
  blockcol = col%block_size;

  for ( k=1; k<=levels; k=k+1 ) {
    subheight = band_size[k][0];
    subwidth = band_size[k][1];
    pos_row = band_pos[k][0];
    pos_col = band_pos[k][1];
    x_org = (float)(subwidth-1)/2;
    y_org = (float)(subheight-1)/2;
    x0 = x_org + x*pow(0.5, k);
    y0 = y_org + y*pow(0.5, k);
    max_radius = h_size[k-1];
    if (v_size[k-1]>max_radius) {
      max_radius = v_size[k-1];
    }
    if (d_size[k-1]>max_radius) {
      max_radius = d_size[k-1];
    }
    for ( i=(int)y0-max_radius; i<=(int)y0+max_radius; i=i+1 ) {
      for ( j=(int)x0-max_radius; j<=(int)x0+max_radius; j=j+1 ) {
        if ( (i>=0) && (i<subheight) && (j>=0) && (j<subwidth) ) {
          ExtWAC(WAC, k, i, j, band_size);
        }
      }
    }
    for ( i=(int)y0-h_size[k-1]; i<=(int)y0+h_size[k-1]; i=i+1 ) {
      for ( j=(int)x0-h_size[k-1]; j<=(int)x0+h_size[k-1]; j=j+1 ) {
	if ( (ZeroWaveletCoef(k, i, j, x, y, h_size[k-1], x_org, y_org)==1) && i>=0 && i<subheight 
	    && j>=0 && j<subwidth ) {
	  value = (WAC[k-1][2*i][2*j]-WAC[k-1][2*i][2*j+1]+
				     WAC[k-1][2*i+1][2*j]-WAC[k-1][2*i+1][2*j+1])/2;
	  addentry(&Smatrix[i][j+pos_col][blockrow], blockcol, value*scaling[3*(levels-k)+1]);

	}
      }
    }
    for ( i=(int)y0-v_size[k-1]; i<=(int)y0+v_size[k-1]; i=i+1 ) {
      for ( j=(int)x0-v_size[k-1]; j<=(int)x0+v_size[k-1]; j=j+1 ) {
	if ( (ZeroWaveletCoef(k, i, j, x, y, v_size[k-1], x_org, y_org)==1) && i>=0 && i<subheight 
	    && j>=0 && j<subwidth) {
	  value = (WAC[k-1][2*i][2*j]+WAC[k-1][2*i][2*j+1]-
				      WAC[k-1][2*i+1][2*j]-WAC[k-1][2*i+1][2*j+1])/2;
	  addentry(&Smatrix[i+pos_row][j][blockrow], blockcol, value*scaling[3*(levels-k)+2]);
	}
      }
    }
    for ( i=(int)y0-d_size[k-1]; i<=(int)y0+d_size[k-1]; i=i+1 ) {
      for ( j=(int)x0-d_size[k-1]; j<=(int)x0+d_size[k-1]; j=j+1 ) {
	if ( (ZeroWaveletCoef(k, i, j, x, y, d_size[k-1], x_org, y_org)==1) && i>=0 && i<subheight 
	     && j>=0 && j<subwidth) {
	  value = (WAC[k-1][2*i][2*j]-WAC[k-1][2*i][2*j+1] -
					       WAC[k-1][2*i+1][2*j]+WAC[k-1][2*i+1][2*j+1])/2;
	  addentry(&Smatrix[i+pos_row][j+pos_col][blockrow], blockcol, value*scaling[3*(levels-k+1)]);
	}
      }
    }
  }

  for ( i=0; i<band_size[levels][0]; i=i+1 ) {
    for ( j=0; j<band_size[levels][1]; j=j+1 ) {
      if ( ZeroWaveletCoef(levels, i, j, x, y, a_size, x_org, y_org) ) {
	addentry(&Smatrix[i][j][blockrow], blockcol, WAC[levels][i][j]*scaling[0]);
      }
    }
  }

}
*/


/***** fast method for computing wavelet transform and store it in new sparse form ******/
void ComputeSparseWT(float ***WAC, MLIST *Smatrix, int a_size, int *h_size, int *v_size, 
		     int *d_size, int levels, float x, float y, int row, int col, int block_size, 
                     float *scaling, int **band_size, int **band_pos)
{
  int i, j, k;
  int subheight, subwidth, blockrow, blockcol, pos_row, pos_col, max_radius;
  float x_org, y_org, x0, y0;
  float value;

  blockrow = row%block_size;
  blockcol = col%block_size;

  for ( k=1; k<=levels; k=k+1 ) {
    subheight = band_size[k][0];
    subwidth = band_size[k][1];
    pos_row = band_pos[k][0];
    pos_col = band_pos[k][1];
    x_org = (float)(subwidth-1)/2;
    y_org = (float)(subheight-1)/2;
    x0 = x_org + x*pow(0.5, k);
    y0 = y_org + y*pow(0.5, k);
    max_radius = h_size[k-1];
    if (v_size[k-1]>max_radius) {
      max_radius = v_size[k-1];
    }
    if (d_size[k-1]>max_radius) {
      max_radius = d_size[k-1];
    }
    for ( i=(int)y0-max_radius; i<=(int)y0+max_radius; i=i+1 ) {
      for ( j=(int)x0-max_radius; j<=(int)x0+max_radius; j=j+1 ) {
        if ( (i>=0) && (i<subheight) && (j>=0) && (j<subwidth) ) {
          ExtWAC(WAC, k, i, j, band_size);
        }
      }
    }
    for ( i=(int)y0-h_size[k-1]; i<=(int)y0+h_size[k-1]; i=i+1 ) {
      for ( j=(int)x0-h_size[k-1]; j<=(int)x0+h_size[k-1]; j=j+1 ) {
	if ( (ZeroWaveletCoef(k, i, j, x, y, h_size[k-1], x_org, y_org)==1) && i>=0 && i<subheight 
	    && j>=0 && j<subwidth ) {
	  value = (WAC[k-1][2*i][2*j]-WAC[k-1][2*i][2*j+1]+
				     WAC[k-1][2*i+1][2*j]-WAC[k-1][2*i+1][2*j+1])/2;
	  addmatrixentry(&Smatrix[i], j+pos_col, blockrow, blockcol, 
			 value*scaling[3*(levels-k)+1]);

	}
      }
    }
    for ( i=(int)y0-v_size[k-1]; i<=(int)y0+v_size[k-1]; i=i+1 ) {
      for ( j=(int)x0-v_size[k-1]; j<=(int)x0+v_size[k-1]; j=j+1 ) {
	if ( (ZeroWaveletCoef(k, i, j, x, y, v_size[k-1], x_org, y_org)==1) && i>=0 && i<subheight 
	    && j>=0 && j<subwidth) {
	  value = (WAC[k-1][2*i][2*j]+WAC[k-1][2*i][2*j+1]-
				      WAC[k-1][2*i+1][2*j]-WAC[k-1][2*i+1][2*j+1])/2;
	  addmatrixentry(&Smatrix[i+pos_row], j, blockrow, blockcol, 
		   value*scaling[3*(levels-k)+2]);
	}
      }
    }
    for ( i=(int)y0-d_size[k-1]; i<=(int)y0+d_size[k-1]; i=i+1 ) {
      for ( j=(int)x0-d_size[k-1]; j<=(int)x0+d_size[k-1]; j=j+1 ) {
	if ( (ZeroWaveletCoef(k, i, j, x, y, d_size[k-1], x_org, y_org)==1) && i>=0 && i<subheight 
	     && j>=0 && j<subwidth) {
	  value = (WAC[k-1][2*i][2*j]-WAC[k-1][2*i][2*j+1] -
					       WAC[k-1][2*i+1][2*j]+WAC[k-1][2*i+1][2*j+1])/2;
	  addmatrixentry(&Smatrix[i+pos_row], j+pos_col, blockrow, blockcol, 
		   value*scaling[3*(levels-k+1)]);
	}
      }
    }
  }

  for ( i=0; i<band_size[levels][0]; i=i+1 ) {
    for ( j=0; j<band_size[levels][1]; j=j+1 ) {
      if ( ZeroWaveletCoef(levels, i, j, x, y, a_size, x_org, y_org) ) {
	addmatrixentry(&Smatrix[i], j, blockrow, blockcol, WAC[levels][i][j]*scaling[0]);
      }
    }
  }

}



/* Apply scaling of each band to wavelet transform */
void ApplyScaling(float **matrix, float *scaling_factors, int num_levels, 
                  int **band_size, int **band_pos)
{
  int rlb, rub, clb, cub, level;
  int i, j;
  rlb = 0; rub = band_pos[num_levels][0]; 
  clb = 0; cub = band_pos[num_levels][1];
  for ( i=rlb; i<rub; i++) {
    for ( j=clb; j<cub; j++) {
	matrix[i][j] = matrix[i][j]/scaling_factors[0];
    }
  }
  for ( level=1; level<=num_levels; level++) {
    rlb = 0; rub = band_size[num_levels-level+1][0]; 
    clb = band_pos[num_levels-level+1][1]; cub = band_pos[num_levels-level+1][1]+band_size[num_levels-level+1][1];
    for ( i=rlb; i<rub; i++) {
      for ( j=clb; j<cub; j++) {
          matrix[i][j] = matrix[i][j]/scaling_factors[3*level-2];
      }
    }
    rlb = band_pos[num_levels-level+1][0]; rub = band_pos[num_levels-level+1][0]+band_size[num_levels-level+1][0];
    clb = 0; cub = band_size[num_levels-level+1][1];
    for ( i=rlb; i<rub; i++) {
      for ( j=clb; j<cub; j++) {
         matrix[i][j] = matrix[i][j]/scaling_factors[3*level-1];
      }
    }
    rlb = band_pos[num_levels-level+1][0]; rub = band_pos[num_levels-level+1][0]+band_size[num_levels-level+1][0];
    clb = band_pos[num_levels-level+1][1]; cub = band_pos[num_levels-level+1][1]+band_size[num_levels-level+1][1];
    for ( i=rlb; i<rub; i++) {
      for ( j=clb; j<cub; j++) {
          matrix[i][j] = matrix[i][j]/scaling_factors[3*level];
      }
    }
  }

}

void conv_noext(float *convdata, float *data, float *filter, 
		int filterlen, int filter_origin, int width)
{
  int i, j;
  for ( i=0; i<width; i=i+1 ) {
    for ( j=0; j<filterlen; j=j+1 ) {
      if ( i+j-filter_origin>=0 && i+j-filter_origin<width ) {
	convdata[i+j-filter_origin] = convdata[i+j-filter_origin] + data[i]*filter[j];
      }
    }
  }
}

void reg_downsamp(float *downsamp_data, float *data, int start, 
		  int halfwidth)
{
  int i;
  for ( i=0; i<halfwidth; i=i+1 ) {
    downsamp_data[i] = data[start+i*2];
  }
}

/* symmetric extention */
/*
void ext_conv_downsp(float *data, int length, float *filter, int filter_len, int start, 
                     float *conved)
{
  int i, j;
  float temp;
  int extended_len = length + 2*(filter_len-1);
  float *extended_data;

  extended_data = malloc(sizeof(float)*extended_len);
  for ( i=filter_len-1; i<length+filter_len-1; i=i+1 ) {
    extended_data[i] = data[i-filter_len+1];
  }
  for ( i=0; i<filter_len-1; i=i+1 ) {
    extended_data[filter_len-2-i] = data[i];
    extended_data[filter_len-1+length+i] = data[length-1-i];
  }
  for ( i=filter_len-1+start; i<extended_len; i=i+2 ) {
    temp = 0;
    for ( j=0; j<filter_len; j=j+1 ) {
      temp = temp + extended_data[i+j-filter_len+1]*filter[j];
    }
    conved[(i-filter_len+1-start)/2] = temp;
  }
  free(extended_data);
  
}
*/

/* zero extention */
void ext_conv_downsp(float *data, int length, float *filter, int filter_len, int start,
                     float *conved)
{
  int i, j;
  float temp;
  int extended_len = length + 2*(filter_len-1);
  float *extended_data;

  extended_data = malloc(sizeof(float)*extended_len);
  for ( i=filter_len-1; i<length+filter_len-1; i=i+1 ) {
    extended_data[i] = data[i-filter_len+1];
  }
  for ( i=0; i<filter_len-1; i=i+1 ) {
    extended_data[filter_len-2-i] = 0;
    extended_data[filter_len-1+length+i] = 0;
  }
  for ( i=filter_len-1+start; i<extended_len; i=i+2 ) {
    temp = 0;
    for ( j=0; j<filter_len; j=j+1 ) {
      temp = temp + extended_data[i+j-filter_len+1]*filter[j];
    }
    conved[(i-filter_len+1-start)/2] = temp;
  }
  free(extended_data);

}


/* this haar transform is supposed to handle all kinds of image sizes */
void new_haar_transform(float **data, int levels, int start, 
                        int **band_size, int **band_pos, float **wave_coef)
{
  int i, j, n;
  float **l, **h, **tempdata;
  float *tempvec, *tempconved;
  float lpfilter[2] = {1/sqrt(2), 1/sqrt(2)};
  float hpfilter[2] = {1/sqrt(2), -1/sqrt(2)};

  tempdata = multialloc(sizeof(float), 2, band_size[0][0], band_size[0][1]);
  for ( i=0; i<band_size[0][0]; i=i+1 ) {
    for ( j=0; j<band_size[0][1]; j=j+1 ) {
      tempdata[i][j] = data[i][j];
    }
  }
  
  for ( n=0; n<levels; n=n+1 ) {
    l = multialloc(sizeof(float), 2, band_size[n][0], band_size[n+1][1]);
    h = multialloc(sizeof(float), 2, band_size[n][0], band_size[n+1][1]);
    for ( i=0; i<band_size[n][0]; i=i+1 ) {
      ext_conv_downsp(tempdata[i], band_size[n][1], lpfilter, 2, start, l[i]);
      ext_conv_downsp(tempdata[i], band_size[n][1], hpfilter, 2, start, h[i]);
    }
    multifree(tempdata, 2);
    tempdata = multialloc(sizeof(float), 2, band_size[n+1][0], band_size[n+1][1]);
    tempvec = malloc(sizeof(float)*band_size[n][0]);
    tempconved = malloc(sizeof(float)*band_size[n+1][0]);
    for ( j=0; j<band_size[n+1][1]; j=j+1 ) {
      for ( i=0; i<band_size[n][0]; i=i+1 ) {
        tempvec[i] = l[i][j];
      }
      ext_conv_downsp(tempvec, band_size[n][0], lpfilter, 2, start, tempconved);
      for ( i=0; i<band_size[n+1][0]; i=i+1 ) {
        tempdata[i][j] = tempconved[i];
      }
      ext_conv_downsp(tempvec, band_size[n][0], hpfilter, 2, start, tempconved);
      for ( i=0; i<band_size[n+1][0]; i=i+1 ) {
        wave_coef[band_pos[n+1][0]+i][j] = tempconved[i];
      }
      for ( i=0; i<band_size[n][0]; i=i+1 ) {
        tempvec[i] = h[i][j];
      }
      ext_conv_downsp(tempvec, band_size[n][0], lpfilter, 2, start, tempconved);
      for ( i=0; i<band_size[n+1][0]; i=i+1 ) {
        wave_coef[i][band_pos[n+1][1]+j] = tempconved[i];
      }
      ext_conv_downsp(tempvec, band_size[n][0], hpfilter, 2, start, tempconved);
      for ( i=0; i<band_size[n+1][0]; i=i+1 ) {
        wave_coef[band_pos[n+1][0]+i][band_pos[n+1][1]+j] = tempconved[i];
      }
    }
    multifree(h, 2);
    multifree(l, 2);
    free(tempvec);
    free(tempconved);
  }
  for ( i=0; i<band_size[levels][0]; i=i+1 ) {
    for ( j=0; j<band_size[levels][1]; j=j+1 ) {
      wave_coef[i][j] = tempdata[i][j];
    }
  }
  multifree(tempdata, 2);
}


/* This Haar wavelet transform only handles image size of power of 2 */
void haar_transform(float **data, int height, int width, int levels, int start)
{
  int i, j, n;
  int subwidth, subheight;
  float **l;
  float **h;
  float **tempdata;
  float *conved;
  float *temp;
  float lpfilter[2] = {1/sqrt(2), 1/sqrt(2)};
  float hpfilter[2] = {-1/sqrt(2), 1/sqrt(2)};
  int filterlen_l = 2;
  int filterlen_h = 2;
  int lp_origin = 0;
  int hp_origin = 0;

  for (n=0; n<levels; n=n+1 ) {
    subheight = height>>n;
    subwidth = width>>n;
    tempdata = multialloc(sizeof(float), 2, subheight, subwidth);
    for ( i=0; i<subheight; i=i+1 ) {
      for ( j=0; j<subwidth; j=j+1 ) {
	tempdata[i][j] = data[i][j];
      }
    }

    l = multialloc(sizeof(float), 2, subheight, subwidth/2);
    h = multialloc(sizeof(float), 2, subheight, subwidth/2);
    conved = malloc(sizeof(float)*subwidth);
    for ( i=0; i<subheight; i=i+1 ) {
      clear_vector(conved, subwidth);
      conv_noext(conved, tempdata[i], lpfilter, filterlen_l, lp_origin, subwidth);
      for ( j=0; j<subwidth/2; j=j+1 ) {
	l[i][j] = conved[start+j*2];
      }
    }
    for ( i=0; i<subheight; i=i+1 ) {
      clear_vector(conved, subwidth);
      conv_noext(conved, tempdata[i], hpfilter, filterlen_h, hp_origin, subwidth);
      for ( j=0; j<subwidth/2; j=j+1 ) {
	h[i][j] = conved[start+j*2];
      }
    }
    free(conved);
    conved = malloc(sizeof(float)*subheight);
    temp = malloc(sizeof(float)*subheight);
    for ( j=0; j<subwidth/2; j=j+1 ) {
      clear_vector(conved, subheight);
      for ( i=0; i<subheight; i=i+1 ) {
	temp[i] = l[i][j];
      }
      conv_noext(conved, temp, lpfilter, filterlen_l, lp_origin, subheight);
      for ( i=0; i<subheight/2; i=i+1 ) {
	data[i][j] = conved[start+i*2];
      }
      clear_vector(conved, subheight);
      conv_noext(conved, temp, hpfilter, filterlen_h, hp_origin, subheight);
      for ( i=0; i<subheight/2; i=i+1 ) {
	data[i+subheight/2][j] = conved[start+i*2];
      }
    }
    for ( j=0; j<subwidth/2; j=j+1 ) {
      clear_vector(conved, subheight);
      for ( i=0; i<subheight; i=i+1 ) {
	temp[i] = h[i][j];
      }
      conv_noext(conved, temp, lpfilter, filterlen_l, lp_origin, subheight);
      for ( i=0; i<subheight/2; i=i+1 ) {
	data[i][j+subwidth/2] = conved[start+i*2];
      }
      clear_vector(conved, subheight);
      conv_noext(conved, temp, hpfilter, filterlen_h, hp_origin, subheight);
      for ( i=0; i<subheight/2; i=i+1 ) {
	data[i+subheight/2][j+subwidth/2] = conved[start+i*2];
      }
    }
    free(conved);
    free(temp);
    multifree(tempdata, 2);
    multifree(l, 2);
    multifree(h, 2);
  }
}


/* Computing convolved image using lossy coded BIG matrix */
void DecodeMultiply(LIST ** listmatrix, int height_conved, int width_conved, int height_padded, 
                    int width_padded, int **band_size, int **band_pos, float ** img1, 
		    float ** img3, int num_levels, float delta, float *scaling_factors)
{
  int i, j;
  NODE* ptr;
  float **wave_coef;
  clock_t time_1, time_2;
  double cputime;

  wave_coef = multialloc(sizeof(float), 2, height_conved, width_conved);

  time_1 = clock();
  new_haar_transform(img1, num_levels, 1, band_size, band_pos, wave_coef);
  time_2 = clock();
  cputime = (double) (time_2-time_1)/(double)CLOCKS_PER_SEC;
  printf("cpu time for doing haar transform = %f\n", cputime);
  
  ApplyScaling(wave_coef, scaling_factors, num_levels, band_size, band_pos);

  time_1 = clock();
  for ( i=0; i<height_padded; i++) {
    for ( j=0; j<width_padded; j++) {
      img3[i][j] = 0;
      ptr = listmatrix[i][j].head;
      while ( ptr != NULL) {
	img3[i][j] += wave_coef[ptr->pixel_row][ptr->pixel_col]*(float)ptr->pixel_val;
	ptr = ptr->link;
      }
    }
  }
  time_2 = clock();
  cputime = (double) (time_2-time_1)/(double)CLOCKS_PER_SEC;
  printf("cpu time for computing matrix vector multiply = %f\n", cputime);

  for ( i=0; i<height_padded; i++) {
    for ( j=0; j<width_padded; j++) {
      img3[i][j] = img3[i][j]*delta;
    }
  }

  multifree(wave_coef, 2);

}


void BlockInverseTransform(float **img3, int height_padded, int width_padded, int block_size, 
			   int num_levels, float *lprfilter, float *hprfilter)
{
  int i, j, m, n;
  float **block_img;
  block_img = multialloc(sizeof(float), 2, block_size, block_size);
  for ( m=0; m<height_padded/block_size; m=m+1) {
    for ( n=0; n<width_padded/block_size; n=n+1) {
      clear_matrix(block_img, block_size, block_size);
      for ( i=m*block_size; i<(m+1)*block_size; i++) {
	for ( j=n*block_size; j<(n+1)*block_size; j++) {
	  block_img[i-m*block_size][j-n*block_size] = img3[i][j];
	}
      }
      haar_wave_recon(block_img, lprfilter, hprfilter,FILTERLEN_L, 
		 FILTERLEN_H, block_size, block_size, num_levels);
      for ( i=m*block_size; i<(m+1)*block_size; i++) {
	for ( j=n*block_size; j<(n+1)*block_size; j++) {
	  img3[i][j] = block_img[i-m*block_size][j-n*block_size];
	}
      }
    }
  }
  multifree(block_img, 2);
}



float ComputeNRMSE(float **img2, float **img3, int **imgentry, 
		   int imgnonzero)
{
  int i, j, n;
  float energy, sqdiff, nrmse;
  energy = 0;
  sqdiff = 0;
  for ( n=0; n<imgnonzero; n=n+1 ) {
    i = imgentry[n][0];
    j = imgentry[n][1];
    energy += img2[i][j]*img2[i][j];
    sqdiff += (img2[i][j]-img3[i][j])*(img2[i][j]-img3[i][j]);
  }
  nrmse = sqrt(sqdiff/energy);
  return nrmse;
}

void free_matrix(float ** matrix, int height)
{
  int i;
  for ( i=0; i<height; i++) {
    free(matrix[i]);
  }
}

int RoundTZero(float num)
{
  int intnum;
  if (num<0) {
    intnum = ceil(num);
    return intnum;
  }
  else {
    intnum = floor(num);
    return intnum;
  }
}

/***** record listmatrix *****/
void RecordListmatrix(FILE *fp, LIST ** listmatrix, 
		      int height, int width)
{
  int i, j;
  NODE * pnode;
  
  for ( i=0; i<height; i=i+1) {
    for ( j=0; j<width; j=j+1) {

	fprintf(fp, "%d\n", listmatrix[i][j].num_pix);
        for ( pnode = listmatrix[i][j].head; pnode!=NULL; pnode = pnode->link) {
	  fprintf(fp, "%d\n%d\n", pnode->pixel_row, pnode->pixel_col);
	  fprintf(fp, "%d\n", pnode->pixel_val);
        }
    }
  }
}

void ReadListmatrix(FILE *fp, LIST ** listmatrix, int height, int width)
{
  int i, j, n;
  int pix_row, pix_col, pix_val, num_pix;
  for ( i=0; i<height; i=i+1) {
    for ( j=0; j<width; j=j+1) {
      fscanf(fp, "%d", &num_pix);
      for ( n=0; n<num_pix; n=n+1) {
	fscanf(fp, "%d", &pix_row);
	fscanf(fp, "%d", &pix_col);
	fscanf(fp, "%d", &pix_val);
	AddItem(&listmatrix[i][j], pix_row, pix_col, pix_val);
      }
    }
  }
}

void OldApplyScaling(float **matrix, float *scaling_factors, 
                  int num_levels, int height, int width)
{
  int rlb, rub, clb, cub, level;
  int i, j;
  rlb = 0; rub = (height>>num_levels); 
  clb = 0; cub = (width>>num_levels);
  for ( i=rlb; i<rub; i++) {
    for ( j=clb; j<cub; j++) {
	matrix[i][j] = matrix[i][j]/scaling_factors[0];
    }
  }
  for ( level=1; level<=num_levels; level++) {
    rlb = 0; rub = (height>>(num_levels-level+1)); 
    clb = (width>>(num_levels-level+1)); cub = (width>>(num_levels-level));
    for ( i=rlb; i<rub; i++) {
      for ( j=clb; j<cub; j++) {
          matrix[i][j] = matrix[i][j]/scaling_factors[3*level-2];
      }
    }
    rlb = (height>>(num_levels-level+1)); rub = (height>>(num_levels-level));
    clb = 0; cub = (width>>(num_levels-level+1));
    for ( i=rlb; i<rub; i++) {
      for ( j=clb; j<cub; j++) {
         matrix[i][j] = matrix[i][j]/scaling_factors[3*level-1];
      }
    }
    rlb = (height>>(num_levels-level+1)); rub = (height>>(num_levels-level));
    clb = (width>>(num_levels-level+1)); cub = (width>>(num_levels-level));
    for ( i=rlb; i<rub; i++) {
      for ( j=clb; j<cub; j++) {
          matrix[i][j] = matrix[i][j]/scaling_factors[3*level];
      }
    }
  }

}

float OldFastComputeApproxCoef(int k, int m, int n, float ***WAC, PARAMETER para, 
			       struct power_LUT *powertable, float x, float y, 
			       int height, int width, int *h_size, 
			       int *v_size, int *d_size, float pixel_size)
{
  float value = 0;
  float x_org, y_org, x_suborg, y_suborg;
  int maxradius;

  if (k==0) {
    x_org = (float)(width-1)/2;
    y_org = (float)(height-1)/2;
    value = computepsf(para, powertable, (float)n-x_org, (float)m-y_org, x, y, pixel_size);
    WAC[k][m][n] = value;
    return value;
  }

  maxradius = h_size[k-1];
  if (maxradius<v_size[k-1]) {
    maxradius = v_size[k-1];
  }
  if (maxradius<d_size[k-1]) {
    maxradius = d_size[k-1];
  }
  x_suborg = (float)((width>>k)-1)/2;
  y_suborg = (float)((height>>k)-1)/2;
  if (ZeroWaveletCoef(k, m, n, x, y, maxradius, x_suborg, y_suborg)==0) {
    x_org = (float)(width-1)/2;
    y_org = (float)(height-1)/2;
    value = computepsf(para, powertable, (float)(n<<k)-x_org, (float)(m<<k)-y_org, x, y, pixel_size);
    value = value*pow(2,k);
    WAC[k][m][n] = value;
    return value;
  }
  else {
    value = value + OldFastComputeApproxCoef(k-1, 2*m, 2*n, WAC, para, powertable, 
					  x, y, height, width, h_size, v_size, d_size, pixel_size);
    value = value + OldFastComputeApproxCoef(k-1, 2*m, 2*n+1, WAC, para, powertable, 
					  x, y, height, width, h_size, v_size, d_size, pixel_size);
    value = value + OldFastComputeApproxCoef(k-1, 2*m+1, 2*n, WAC, para, powertable, 
					  x, y, height, width, h_size, v_size, d_size, pixel_size);
    value = value + OldFastComputeApproxCoef(k-1, 2*m+1, 2*n+1, WAC, para, powertable, 
					  x, y, height, width, h_size, v_size, d_size, pixel_size);
    value = value/2;
    WAC[k][m][n] = value;
    return value;
  }

}

/***** fast method for computing wavelet transform and store it in sparse form ******/
void OldComputeSparseWT(float ***WAC, LINKLIST ***Smatrix, int *h_size, int *v_size, int *d_size, 
		     int height, int width, int levels, int row, int col, int block_size, 
		     float *scaling)
{
  int i, j, k;
  int subheight, subwidth, blockrow, blockcol;
  float x_org, y_org, x0, y0, x, y;
  float value;

  x_org = (float)(width-1)/2;
  y_org = (float)(height-1)/2;
  x = (float)col-x_org;
  y = (float)row-y_org;
  blockrow = row%block_size;
  blockcol = col%block_size;

  for ( k=1; k<=levels; k=k+1 ) {
    subheight = (height>>k);
    subwidth = (width>>k);
    x_org = (float)(subwidth-1)/2;
    y_org = (float)(subheight-1)/2;
    x0 = x_org + x*pow(0.5, k);
    y0 = y_org + y*pow(0.5, k);
    for ( i=(int)y0-h_size[k-1]; i<=(int)y0+h_size[k-1]; i=i+1 ) {
      for ( j=(int)x0-h_size[k-1]; j<=(int)x0+h_size[k-1]; j=j+1 ) {
	if ( (ZeroWaveletCoef(k, i, j, x, y, h_size[k-1], x_org, y_org)==1) && i>=0 && i<subheight 
	    && j>=0 && j<subwidth ) {
	  value = (WAC[k-1][2*i][2*j]-WAC[k-1][2*i][2*j+1]+
				     WAC[k-1][2*i+1][2*j]-WAC[k-1][2*i+1][2*j+1])/2;
	  addentry(&Smatrix[i][j+subwidth][blockrow], blockcol, value*scaling[3*(levels-k)+1]);

	}
      }
    }
    for ( i=(int)y0-v_size[k-1]; i<=(int)y0+v_size[k-1]; i=i+1 ) {
      for ( j=(int)x0-v_size[k-1]; j<=(int)x0+v_size[k-1]; j=j+1 ) {
	if ( (ZeroWaveletCoef(k, i, j, x, y, v_size[k-1], x_org, y_org)==1) && i>=0 && i<subheight 
	    && j>=0 && j<subwidth) {
	  value = (WAC[k-1][2*i][2*j]+WAC[k-1][2*i][2*j+1]-
				      WAC[k-1][2*i+1][2*j]-WAC[k-1][2*i+1][2*j+1])/2;
	  addentry(&Smatrix[i+subheight][j][blockrow], blockcol, value*scaling[3*(levels-k)+2]);
	}
      }
    }
    for ( i=(int)y0-d_size[k-1]; i<=(int)y0+d_size[k-1]; i=i+1 ) {
      for ( j=(int)x0-d_size[k-1]; j<=(int)x0+d_size[k-1]; j=j+1 ) {
	if ( (ZeroWaveletCoef(k, i, j, x, y, d_size[k-1], x_org, y_org)==1) && i>=0 && i<subheight 
	     && j>=0 && j<subwidth) {
	  value = (WAC[k-1][2*i][2*j]-WAC[k-1][2*i][2*j+1] -
					       WAC[k-1][2*i+1][2*j]+WAC[k-1][2*i+1][2*j+1])/2;
	  addentry(&Smatrix[i+subheight][j+subwidth][blockrow], blockcol, value*scaling[3*(levels-k+1)]);
	}
      }
    }
  }
  addentry(&Smatrix[0][0][blockrow], blockcol, WAC[levels][0][0]*scaling[0]);

}

void Copy2SparseColumn(MLIST *Smatrix, LINKLIST *sparse_column, int column_i, 
		       int column_j)
{
  MENTRY *ptr = Smatrix[column_i].head;
  while (ptr!=NULL) {
    if (ptr->column_j==column_j) {
      addentry(&sparse_column[ptr->row_i], ptr->row_j, ptr->pixel_val);
    }
    ptr = ptr->link;
  }
}

void IndexPixel2Check(int **index, float upleft_x, float upleft_y, float lowright_x, 
		      float lowright_y, int *max_radius_array, int **band_pos, int **band_size, int levels)
{
  int k;
  int row_lb, row_ub, col_lb, col_ub;
  for ( k=1; k<=levels; k++ ) {
    row_lb = floor(upleft_y*pow(0.5, (double)k) + (float)(band_size[k][0]-1)/2)-max_radius_array[k-1];
    col_lb = floor(upleft_x*pow(0.5, (double)k) + (float)(band_size[k][1]-1)/2)-max_radius_array[k-1];
    row_ub = ceil(lowright_y*pow(0.5, (double)k) + (float)(band_size[k][0]-1)/2)+max_radius_array[k-1];
    col_ub = ceil(lowright_x*pow(0.5, (double)k) + (float)(band_size[k][1]-1)/2)+max_radius_array[k-1];
    if (row_lb<0) {
      row_lb = 0;
    }
    if (row_ub>=band_size[k][0]) {
      row_ub = band_size[k][0]-1;
    }
    if (col_lb<0) {
      col_lb = 0;
    }
    if (col_ub>=band_size[k][1]) {
      col_ub = band_size[k][1]-1;
    }
    index[3*(k-1)][0] = row_lb;
    index[3*(k-1)][1] = row_ub;
    index[3*(k-1)][2] = col_lb+band_pos[k][1];
    index[3*(k-1)][3] = col_ub+band_pos[k][1];
    index[3*(k-1)+1][0] = row_lb+band_pos[k][0];
    index[3*(k-1)+1][1] = row_ub+band_pos[k][0];
    index[3*(k-1)+1][2] = col_lb;
    index[3*(k-1)+1][3] = col_ub;
    index[3*(k-1)+2][0] = row_lb+band_pos[k][0];
    index[3*(k-1)+2][1] = row_ub+band_pos[k][0];
    index[3*(k-1)+2][2] = col_lb+band_pos[k][1];
    index[3*(k-1)+2][3] = col_ub+band_pos[k][1];
  }
  index[3*levels][0] = row_lb;
  index[3*levels][1] = row_ub;
  index[3*levels][2] = col_lb;
  index[3*levels][3] = col_ub;
}


/***********************************************/
/*********** This is MAIN function *************/
/***********************************************/

int main(int argc, char *argv[])
{
  int i, j, k, l, m, n, h, w, temp, non0entry;
  int num_levels, height, width, block_size, block_num_levels;
  int height_conved, width_conved, height_padded, width_padded;
  int tempval, imgnonzero;
  int *h_size, *v_size, *d_size, *max_radius_array;
  int a_size;
  float delta; /* quantization interval */
  float error, x_org, y_org, x, y, pixel_size;
  float **img1, **img2, **img3;
  int **imgentry; /* randomly sampled locations for computing NRMSE */
  int **index;
  LINKLIST *block_wavecoef;
  LINKLIST *sparse_column;
  float *scaling_factors, *inverse_scaling;
  /* img1 is input image, img2 is result of rigorous conv, img3 is 
     conv result of lossy matrix coding*/
  FILE *fp, *fq;
  struct parameter para;
  struct power_LUT *powertable;
  clock_t time_1, time_2, time_3, time_4; 
  double cputime;
  char outputfile[50], imagefile[50], scalefile[50], convedfile[50];
  char matrixsavefile[50], bandinfo[50], imginfo[50];
  LIST ** listmatrix;
  MLIST *Smatrix;
  float ***WAC;
  int *** flag_matrix;
  int **band_size;
  int **band_pos;

  float lpfilter[2] = {1/sqrt(2), 1/sqrt(2)};
  float hpfilter[2] = {1/sqrt(2), -1/sqrt(2)};
  float lprfilter[2] = {1/sqrt(2), 1/sqrt(2)};
  float hprfilter[2] = {-1/sqrt(2), 1/sqrt(2)};
  ENTRY *ptr;
  

  if ( (fp = fopen(argv[1], "r")) ) {
    fscanf(fp, "%f", &(para.a));
    fscanf(fp, "%f", &(para.b));
    fscanf(fp, "%f", &(para.f));
    fscanf(fp, "%f", &(para.g));
    fscanf(fp, "%f", &delta);
    fscanf(fp, "%d", &num_levels);
    fscanf(fp, "%d", &block_num_levels);
    fscanf(fp, "%d", &height);
    fscanf(fp, "%d", &width);
    fscanf(fp, "%d", &block_size);
    fscanf(fp, "%s", outputfile);/* ground truth */
    fscanf(fp, "%s", convedfile);/* lossy coding convolution result */
    fscanf(fp, "%s", imagefile);
    fscanf(fp, "%s", scalefile);
    fscanf(fp, "%s", matrixsavefile);
    fscanf(fp, "%s", bandinfo);
    fscanf(fp, "%s", imginfo);
    fclose(fp);
  }
  else {
    printf("cannot open parameter file\n");
    return 1;
  }
  para.a = para.a*1e-6;
  para.b = para.b*1e-6;
  x_org = (float)(width-1)/2;
  y_org = (float)(height-1)/2;
  pixel_size = IMG_HEIGHT/(float)height;
  height_padded = ceil((double)height/block_size)*block_size;
  width_padded = ceil((double)width/block_size)*block_size;
  printf("delta = %f\n", delta);
  printf("Parameters: a=%f, b=%f, f=%f, g=%f\n", para.a,para.b,
	 para.f,para.g);
  printf("Number of levels for block transform %d\n", block_num_levels);
  printf("Write output file to %s\n", convedfile);
  printf("padded height = %d, padded width = %d\n", height_padded, width_padded);

  h_size = malloc(sizeof(int)*num_levels);
  v_size = malloc(sizeof(int)*num_levels);
  d_size = malloc(sizeof(int)*num_levels);
  max_radius_array = malloc(sizeof(int)*num_levels);
  scaling_factors = malloc(sizeof(float)*(1+3*num_levels));
  inverse_scaling = malloc(sizeof(float)*(1+3*num_levels));
  listmatrix = multialloc(sizeof(LIST), 2, height_padded, width_padded);
  block_wavecoef = malloc(sizeof(LINKLIST)*block_size);
  sparse_column = malloc(sizeof(LINKLIST)*block_size);
  band_size = multialloc(sizeof(int), 2, num_levels+1, 2);
  band_pos = multialloc(sizeof(int), 2, num_levels+1, 2);
  img1 = multialloc(sizeof(float), 2, height, width);
  img2 = multialloc(sizeof(float), 2, height, width);
  index = multialloc(sizeof(int), 2, 3*num_levels+1, 4);

  powertable = build_power_LUT(para.f, MAXVAL, SAMPLES);

  band_size[0][0] = height;
  band_size[0][1] = width;
  for ( i=1; i<=num_levels; i=i+1 ) {
    band_size[i][0] = floor((band_size[i-1][0]+1)/2);
    band_size[i][1] = floor((band_size[i-1][1]+1)/2);
  }
  band_pos[num_levels][0] = band_size[num_levels][0];
  band_pos[num_levels][1] = band_size[num_levels][1];
  for ( i=num_levels-1; i>=0; i=i-1 ) {
    band_pos[i][0] = band_pos[i+1][0]+band_size[i+1][0];
    band_pos[i][1] = band_pos[i+1][1]+band_size[i+1][1];
  }
  height_conved = band_pos[0][0];
  width_conved = band_pos[0][1];
  WAC = multialloc(sizeof(float), 3, num_levels+1, band_size[1][0]*2, band_size[1][1]*2);


  /** Read parameter values from files **/
  if ( (fp = fopen(scalefile, "r")) ) {
    for(i=0; i<1+3*num_levels; i++) {
      fscanf(fp, "%f", &scaling_factors[i]);
      printf("scaling factor %d is %f\n", i, scaling_factors[i]);
      inverse_scaling[i] = 1/scaling_factors[i];
    }
    fclose(fp);
  }
  else {
    printf("cannot open scaling factors file\n");
    exit(1);
  }



  if ( (fq = fopen(bandinfo, "r")) ) {
    for ( i=0; i<num_levels; i=i+1 ) {
      fscanf(fq, "%d", &h_size[i]);
    }
    for ( i=0; i<num_levels; i=i+1 ) {
      fscanf(fq, "%d", &v_size[i]);
    }
    for ( i=0; i<num_levels; i=i+1 ) {
      fscanf(fq, "%d", &d_size[i]);
    }
    fscanf(fq, "%d", &a_size);
    fclose(fq);
    printf("Done reading band size file %s\n", bandinfo);
  }
  else {
    printf("cannot open file %s to read\n", bandinfo);
    exit(1);
  }
  for ( i=0; i<num_levels; i=i+1 ) {
    max_radius_array[i] = h_size[i];
    if (max_radius_array[i]<v_size[i]) {
      max_radius_array[i] = v_size[i];
    }
    if (max_radius_array[i]<d_size[i]) {
      max_radius_array[i] = d_size[i];
    }
  }

  if ( (fq = fopen(imginfo, "r")) ) {
    fscanf(fq, "%d", &imgnonzero);
    imgentry = multialloc(sizeof(float), 2, imgnonzero, 2);
    for ( i=0; i<imgnonzero; i=i+1 ) {
      fscanf(fq, "%d", &imgentry[i][0]);
      fscanf(fq, "%d", &imgentry[i][1]);
    }
    fclose(fq);
    printf("Done reading info about image nonzero entry\n");
  }
  else {
    printf("cannot open file %s to read\n", imginfo);
    exit(1);
  }
  /******** finish reading parameter values ********/

  /* reading input image */
  if ( (fq = fopen(imagefile, "r"))) {
    for ( i=0; i<height; i++) {
      for ( j=0; j<width; j++) {
	fscanf(fq, "%d", &temp);
        img1[i][j] = (float)temp;
      }
    }
    fclose(fq);
  }
  else {
    printf("cannot open image file %s\n", imagefile);
    exit(1);
  }


  /* compute output using conventional approach */
  
  for ( i=0; i<height; i=i+1 ) {
    for ( j=0; j<width; j=j+1 ) {
      img2[i][j] = 0;
    }
  }
  if (GROUND_TRUTH==0) {
    RigoConv(para, powertable,  height, width, img1, img2, imgentry, imgnonzero, pixel_size);
  
    if ( (fq = fopen(outputfile, "w")) ) {
      for ( i=0; i<imgnonzero; i=i+1 ) {
	  fprintf(fq, "%f\n", img2[imgentry[i][0]][imgentry[i][1]]);
      }
      fclose(fq);
      printf("done writing ground truth file %s\n", outputfile);
    }
    else {
      printf("cannot open file %s to write\n", outputfile);
      exit(1);
    }
  }
  else {
    if ( (fq = fopen(outputfile, "r")) ) {
      for ( i=0; i<imgnonzero; i=i+1 ) {
        fscanf(fq, "%f", &img2[imgentry[i][0]][imgentry[i][1]]);
      }
      fclose(fq);
      printf("done reading ground truth file %s\n", outputfile);
    }
    else {
      printf("cannot open file %s to read\n", outputfile);
      exit(1);
    }
  }


  /***** transform and encoding *****/

  /* Initialize the linklists */
  for ( i=0; i<height_padded; i++) {
    for ( j=0; j<width_padded; j++) {
      listmatrix[i][j].head = NULL;
      listmatrix[i][j].tail = NULL;
      listmatrix[i][j].num_pix = 0;
    }
  }

  if (MATRIX_SAVED==0) {
    initial_sparse_matrix(block_wavecoef, block_size);
    initial_sparse_matrix(sparse_column, block_size);
    time_1 = clock();
    Smatrix = malloc(sizeof(MLIST)*height_conved);
    for ( i = 0; i<height_conved; i=i+1 ) {
      Smatrix[i].head = NULL;
      Smatrix[i].tail = NULL;
      Smatrix[i].num_pix = 0;
    }

    for ( m=0; m<height_padded/block_size; m=m+1) {
      for ( n=0; n<width_padded/block_size; n=n+1) {
        time_3 = clock();

        for ( i=m*block_size; i<(m+1)*block_size; i++) {
	  for ( j=n*block_size; j<(n+1)*block_size; j++) {
            if ( (i<height) && (j<width) ) {
	      x = (float)j-x_org;
	      y = (float)i-y_org;
	       
	      for ( k=0; k<band_size[num_levels][0]; k=k+1 ) {
		for ( l=0; l<band_size[num_levels][1]; l=l+1 ) {
		  if ( ZeroWaveletCoef(num_levels, k, l, x, y, a_size, 
				       (float)(band_size[num_levels][1]-1)/2, 
				       (float)(band_size[num_levels][0]-1)/2) ) {
		    FastComputeApproxCoef(num_levels, k, l, WAC, para, powertable, x, y, height, 
					  width, h_size, v_size, d_size,band_size,pixel_size);
		  }
		}
	      }
	      ComputeSparseWT(WAC, Smatrix, a_size, h_size, v_size, d_size, num_levels, 
                              x, y, i, j, block_size, scaling_factors, 
                              band_size, band_pos);
            }
	  }
        } /* finish transform along the rows (full row) for one block*/
	IndexPixel2Check(index, (float)n*block_size-x_org, (float)m*block_size-y_org, 
			 (float)(n+1)*block_size-x_org, (float)(m+1)*block_size-y_org, 
			 max_radius_array, band_pos, band_size, num_levels);        
	for ( j=0; j<3*num_levels+1; j++) {
	  for ( k=index[j][0]; k<=index[j][1]; k++) {
	    for ( l=index[j][2]; l<=index[j][3]; l++) {
	      Copy2SparseColumn(Smatrix, sparse_column, k, l);

	      sparse_haar_decomp(lpfilter, hpfilter, sparse_column, block_wavecoef, 
			    block_size, block_size, block_num_levels, FILTERLEN_L, FILTERLEN_H);
                    
	      for (i=0; i<block_size; i=i+1) {
		ptr = block_wavecoef[i].head;
		while (ptr!=NULL) {
		  tempval = ptr->pixel_val/delta;
		  if (tempval!=0) {
		    AddItem(&listmatrix[i+m*block_size][ptr->pixel_pos+n*block_size], 
			    k, l, tempval);
		  }
		  ptr = ptr->link;
		}
	      }
          
	      clear_sparse_matrix(sparse_column, block_size);
	      clear_sparse_matrix(block_wavecoef, block_size);
	    }
	  }
        }
        /* finish transform one block along the columns (partial column) */

	for ( i=0; i<height_conved; i=i+1 ) {
	  clearmatrixlist(&Smatrix[i]);
	}

              non0entry = 0;
  	      for ( i=0; i<height_padded; i++) {
                for ( j=0; j<width_padded; j++) {
                  non0entry = non0entry + listmatrix[i][j].num_pix;
                }
              }
              printf("non zero matrix entries up to this block = %d\n", non0entry);

        time_4 = clock();
	printf("done block m=%d, n=%d\n", m, n);
	cputime = (double)(time_4-time_3)/(double)CLOCKS_PER_SEC;        
        printf("cpu time for this block is %f\n", cputime);	
      }
    }
    free(Smatrix);
    time_2 = clock();
    cputime = (double) (time_2-time_1)/(double)CLOCKS_PER_SEC;
    printf("CPU time for encoding is %e\n", cputime);  
  
    if ( (fq = fopen(matrixsavefile, "w")) ) {
      RecordListmatrix(fq, listmatrix, height_padded, width_padded);
      printf("successfully written matrixsavefile\n");
      fclose(fq);
    }
    else {
      printf("cannot open matrixsavefile to save matrix\n");
      exit(1);
    }
  }
  else {
    /* when matrix has been saved, use this to read it */
  
    if ( (fq=fopen(matrixsavefile, "r")) ) {
      ReadListmatrix(fq, listmatrix, height_padded, width_padded);
      printf("successfully read listmatrix\n");
      fclose(fq);
    }
    else {
      printf("cannot open matrixsavefile to read\n");
      exit(1);
    }
  }

  free(block_wavecoef);
  free(sparse_column);
  multifree(WAC, 3);

  non0entry = 0;
  for ( i=0; i<height_padded; i++) {
    for ( j=0; j<width_padded; j++) {
      non0entry = non0entry + listmatrix[i][j].num_pix;
    }
  }
  printf("non zero matrix entries = %d\n", non0entry);



  time_1 = clock();
  img3 = multialloc(sizeof(float), 2, height_padded, width_padded);
  DecodeMultiply(listmatrix, height_conved, width_conved, height_padded, width_padded, 
                 band_size, band_pos, img1, img3, num_levels, delta, scaling_factors);
  time_3 = clock();
  
  BlockInverseTransform(img3, height_padded, width_padded, block_size, block_num_levels, 
			lprfilter, hprfilter);
  
  time_4 = clock();
  cputime = (double) (time_4-time_3)/(double)CLOCKS_PER_SEC;
  printf("cpu time for inverse wavelet transform = %f\n", cputime);
  time_2 = clock();
  cputime = (double) (time_2-time_1)/(double)CLOCKS_PER_SEC;
  printf("CPU time for efficient computation is %e\n", cputime);  
  if ( (fq=fopen(convedfile, "w"))) {
    for ( i=0; i<height_padded; i++) {
      for ( j=0; j<width_padded; j++) {
	fprintf(fq, "%f\n", img3[i][j]);
      }
    }
    fclose(fq);
  }
  else {
    printf("cannot open convedfile %s to write\n", convedfile);
  }

  
  error = ComputeNRMSE(img2, img3, imgentry, imgnonzero);
  printf("NRMSE = %f\n", error);
  

  distroy_power_LUT(powertable);
  
  multifree(img1, 2);
  multifree(img2, 2);
  multifree(img3, 2);
  multifree(index, 2);
  
  for ( i=0; i<height_padded; i++) {
    for ( j=0; j<width_padded; j++) {
      ClearQueue(&listmatrix[i][j]);
    }
  }
  multifree(listmatrix,2);
  free(scaling_factors);
  free(inverse_scaling);

  return 0;
}





