/* Created by Jianing Wei */
/* ApplyScaling changed */
/* Dynamic Allocation of WAC changed  */
/* Extention mode needs to be zero extention to be orthonormal */
/* This code also changes the data structure for storing sparse column data */
/* A bug in FastComputeApproxCoef is fixed */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <time.h>

#include "sparseWT.h"
#include "power.h"
#include "list.h"
#include "sparselink.h"
#include "psf.h"

#ifndef FILTERLEN_L
#define FILTERLEN_L 2
#endif
#ifndef FILTERLEN_H
#define FILTERLEN_H 2
#endif


void Write2Dfloat(float **matrix, int height, int width, char *filename)
{
  FILE *fp;
  int i, j;

  if ( fp = fopen(filename, "w") ) {
    for (i=0; i<height; i++) {
      for (j=0; j<width; j++) {
	fprintf(fp, "%f ", matrix[i][j]);
      }
      fprintf(fp, "\n");
    }
    fclose(fp);
  }
  else {
    printf("cannot open file %s to write\n", filename);
    exit(-1);
  }
}


/* Apply scaling of each band to wavelet transform */
void ApplyScaling(float **matrix, float *scaling_factors, int num_levels, 
                  int **band_size, int **band_pos)
{
  int rlb, rub, clb, cub, level;
  int i, j;
  rlb = 0; rub = band_pos[num_levels][0]; 
  clb = 0; cub = band_pos[num_levels][1];
  for ( i=rlb; i<rub; i++) {
    for ( j=clb; j<cub; j++) {
	matrix[i][j] = matrix[i][j]/scaling_factors[0];
    }
  }
  for ( level=1; level<=num_levels; level++) {
    rlb = 0; rub = band_size[num_levels-level+1][0]; 
    clb = band_pos[num_levels-level+1][1]; cub = band_pos[num_levels-level+1][1]+band_size[num_levels-level+1][1];
    for ( i=rlb; i<rub; i++) {
      for ( j=clb; j<cub; j++) {
          matrix[i][j] = matrix[i][j]/scaling_factors[3*level-2];
      }
    }
    rlb = band_pos[num_levels-level+1][0]; rub = band_pos[num_levels-level+1][0]+band_size[num_levels-level+1][0];
    clb = 0; cub = band_size[num_levels-level+1][1];
    for ( i=rlb; i<rub; i++) {
      for ( j=clb; j<cub; j++) {
         matrix[i][j] = matrix[i][j]/scaling_factors[3*level-1];
      }
    }
    rlb = band_pos[num_levels-level+1][0]; rub = band_pos[num_levels-level+1][0]+band_size[num_levels-level+1][0];
    clb = band_pos[num_levels-level+1][1]; cub = band_pos[num_levels-level+1][1]+band_size[num_levels-level+1][1];
    for ( i=rlb; i<rub; i++) {
      for ( j=clb; j<cub; j++) {
          matrix[i][j] = matrix[i][j]/scaling_factors[3*level];
      }
    }
  }

}

/* zero extention */
void ext_conv_downsp(float *data, int length, float *filter, int filter_len, int start,
                     float *conved)
{
  int i, j;
  float temp;
  int extended_len = length + 2*(filter_len-1);
  float *extended_data;

  extended_data = malloc(sizeof(float)*extended_len);
  for ( i=filter_len-1; i<length+filter_len-1; i=i+1 ) {
    extended_data[i] = data[i-filter_len+1];
  }
  for ( i=0; i<filter_len-1; i=i+1 ) {
    extended_data[filter_len-2-i] = 0;
    extended_data[filter_len-1+length+i] = 0;
  }
  for ( i=filter_len-1+start; i<extended_len; i=i+2 ) {
    temp = 0;
    for ( j=0; j<filter_len; j=j+1 ) {
      temp = temp + extended_data[i+j-filter_len+1]*filter[j];
    }
    conved[(i-filter_len+1-start)/2] = temp;
  }
  free(extended_data);

}

void conv_noext(float *convdata, float *data, float *filter, 
		int filterlen, int filter_origin, int width)
{
  int i, j;
  for ( i=0; i<width; i=i+1 ) {
    for ( j=0; j<filterlen; j=j+1 ) {
      if ( i+j-filter_origin>=0 && i+j-filter_origin<width ) {
	convdata[i+j-filter_origin] = convdata[i+j-filter_origin] + data[i]*filter[j];
      }
    }
  }
}


/* this haar transform is supposed to handle all kinds of image sizes */
void new_haar_transform(float **data, int levels, int start, 
                        int **band_size, int **band_pos, float **wave_coef)
{
  int i, j, n;
  float **l, **h, **tempdata;
  float *tempvec, *tempconved;
  float lpfilter[2] = {(float)1/sqrt(2), (float)1/sqrt(2)};
  float hpfilter[2] = {(float)1/sqrt(2), (float)-1/sqrt(2)};

  tempdata = multialloc(sizeof(float), 2, band_size[0][0], band_size[0][1]);
  for ( i=0; i<band_size[0][0]; i=i+1 ) {
    for ( j=0; j<band_size[0][1]; j=j+1 ) {
      tempdata[i][j] = data[i][j];
    }
  }
  
  for ( n=0; n<levels; n=n+1 ) {
    l = multialloc(sizeof(float), 2, band_size[n][0], band_size[n+1][1]);
    h = multialloc(sizeof(float), 2, band_size[n][0], band_size[n+1][1]);
    for ( i=0; i<band_size[n][0]; i=i+1 ) {
      ext_conv_downsp(tempdata[i], band_size[n][1], lpfilter, 2, start, l[i]);
      ext_conv_downsp(tempdata[i], band_size[n][1], hpfilter, 2, start, h[i]);
    }
    multifree(tempdata, 2);
    tempdata = multialloc(sizeof(float), 2, band_size[n+1][0], band_size[n+1][1]);
    tempvec = malloc(sizeof(float)*band_size[n][0]);
    tempconved = malloc(sizeof(float)*band_size[n+1][0]);
    for ( j=0; j<band_size[n+1][1]; j=j+1 ) {
      for ( i=0; i<band_size[n][0]; i=i+1 ) {
        tempvec[i] = l[i][j];
      }
      ext_conv_downsp(tempvec, band_size[n][0], lpfilter, 2, start, tempconved);
      for ( i=0; i<band_size[n+1][0]; i=i+1 ) {
        tempdata[i][j] = tempconved[i];
      }
      ext_conv_downsp(tempvec, band_size[n][0], hpfilter, 2, start, tempconved);
      for ( i=0; i<band_size[n+1][0]; i=i+1 ) {
        wave_coef[band_pos[n+1][0]+i][j] = tempconved[i];
      }
      for ( i=0; i<band_size[n][0]; i=i+1 ) {
        tempvec[i] = h[i][j];
      }
      ext_conv_downsp(tempvec, band_size[n][0], lpfilter, 2, start, tempconved);
      for ( i=0; i<band_size[n+1][0]; i=i+1 ) {
        wave_coef[i][band_pos[n+1][1]+j] = tempconved[i];
      }
      ext_conv_downsp(tempvec, band_size[n][0], hpfilter, 2, start, tempconved);
      for ( i=0; i<band_size[n+1][0]; i=i+1 ) {
        wave_coef[band_pos[n+1][0]+i][band_pos[n+1][1]+j] = tempconved[i];
      }
    }
    multifree(h, 2);
    multifree(l, 2);
    free(tempvec);
    free(tempconved);
  }
  for ( i=0; i<band_size[levels][0]; i=i+1 ) {
    for ( j=0; j<band_size[levels][1]; j=j+1 ) {
      wave_coef[i][j] = tempdata[i][j];
    }
  }
  multifree(tempdata, 2);
}


/* This Haar wavelet transform only handles image size of power of 2 */
void haar_transform(float **data, int height, int width, int levels, int start)
{
  int i, j, n;
  int subwidth, subheight;
  float **l;
  float **h;
  float **tempdata;
  float *conved;
  float *temp;
  float lpfilter[2] = {(float)1/sqrt(2), (float)1/sqrt(2)};
  float hpfilter[2] = {(float)-1/sqrt(2), (float)1/sqrt(2)};
  int filterlen_l = 2;
  int filterlen_h = 2;
  int lp_origin = 0;
  int hp_origin = 0;

  for (n=0; n<levels; n=n+1 ) {
    subheight = height>>n;
    subwidth = width>>n;
    tempdata = multialloc(sizeof(float), 2, subheight, subwidth);
    for ( i=0; i<subheight; i=i+1 ) {
      for ( j=0; j<subwidth; j=j+1 ) {
	tempdata[i][j] = data[i][j];
      }
    }

    l = multialloc(sizeof(float), 2, subheight, subwidth/2);
    h = multialloc(sizeof(float), 2, subheight, subwidth/2);
    conved = malloc(sizeof(float)*subwidth);
    for ( i=0; i<subheight; i=i+1 ) {
      clear_vector(conved, subwidth);
      conv_noext(conved, tempdata[i], lpfilter, filterlen_l, lp_origin, subwidth);
      for ( j=0; j<subwidth/2; j=j+1 ) {
	l[i][j] = conved[start+j*2];
      }
    }
    for ( i=0; i<subheight; i=i+1 ) {
      clear_vector(conved, subwidth);
      conv_noext(conved, tempdata[i], hpfilter, filterlen_h, hp_origin, subwidth);
      for ( j=0; j<subwidth/2; j=j+1 ) {
	h[i][j] = conved[start+j*2];
      }
    }
    free(conved);
    conved = malloc(sizeof(float)*subheight);
    temp = malloc(sizeof(float)*subheight);
    for ( j=0; j<subwidth/2; j=j+1 ) {
      clear_vector(conved, subheight);
      for ( i=0; i<subheight; i=i+1 ) {
	temp[i] = l[i][j];
      }
      conv_noext(conved, temp, lpfilter, filterlen_l, lp_origin, subheight);
      for ( i=0; i<subheight/2; i=i+1 ) {
	data[i][j] = conved[start+i*2];
      }
      clear_vector(conved, subheight);
      conv_noext(conved, temp, hpfilter, filterlen_h, hp_origin, subheight);
      for ( i=0; i<subheight/2; i=i+1 ) {
	data[i+subheight/2][j] = conved[start+i*2];
      }
    }
    for ( j=0; j<subwidth/2; j=j+1 ) {
      clear_vector(conved, subheight);
      for ( i=0; i<subheight; i=i+1 ) {
	temp[i] = h[i][j];
      }
      conv_noext(conved, temp, lpfilter, filterlen_l, lp_origin, subheight);
      for ( i=0; i<subheight/2; i=i+1 ) {
	data[i][j+subwidth/2] = conved[start+i*2];
      }
      clear_vector(conved, subheight);
      conv_noext(conved, temp, hpfilter, filterlen_h, hp_origin, subheight);
      for ( i=0; i<subheight/2; i=i+1 ) {
	data[i+subheight/2][j+subwidth/2] = conved[start+i*2];
      }
    }
    free(conved);
    free(temp);
    multifree(tempdata, 2);
    multifree(l, 2);
    multifree(h, 2);
  }
}


/* Computing convolved image using lossy coded BIG matrix */
void DecodeMultiply(LIST ** listmatrix, int height_conved, int width_conved, int height_padded, 
                    int width_padded, int **band_size, int **band_pos, float ** img1, 
		    float ** img3, int num_levels, float *scaling_factors)
{
  int i, j;
  NODE* ptr;
  float **wave_coef;
  clock_t time_1, time_2;
  double cputime;

  wave_coef = multialloc(sizeof(float), 2, height_conved, width_conved);
  for (i=0; i<height_conved; i++) {
    for (j=0; j<width_conved; j++) {
      wave_coef[i][j] = 0;
    }
  } /* JW: 11/30/2012 initialization */

  time_1 = clock();
  new_haar_transform(img1, num_levels, 1, band_size, band_pos, wave_coef);
  time_2 = clock();
  cputime = (double) (time_2-time_1)/(double)CLOCKS_PER_SEC;
  printf("cpu time for doing haar transform = %f\n", cputime);
  
  ApplyScaling(wave_coef, scaling_factors, num_levels, band_size, band_pos);

  time_1 = clock();
  for ( i=0; i<height_padded; i++) {
    for ( j=0; j<width_padded; j++) {
      img3[i][j] = 0;
      ptr = listmatrix[i][j].head;
      while ( ptr != NULL) {
	img3[i][j] += wave_coef[ptr->pixel_row][ptr->pixel_col]*(float)ptr->pixel_val;
	ptr = ptr->link;
      }
    }
  }
  time_2 = clock();
  cputime = (double) (time_2-time_1)/(double)CLOCKS_PER_SEC;
  printf("cpu time for computing matrix vector multiply = %f\n", cputime);

  /*
  for ( i=0; i<height_padded; i++) {
    for ( j=0; j<width_padded; j++) {
      img3[i][j] = img3[i][j]*delta;
    }
  }
  */

  multifree(wave_coef, 2);

}


static void BlockInverseTransform(float **img3, int height_padded, int width_padded, int block_size, 
                                  int num_levels, float *lprfilter, float *hprfilter)
{
  int i, j, m, n;
  float **block_img;
  block_img = multialloc(sizeof(float), 2, block_size, block_size);
  for ( m=0; m<height_padded/block_size; m=m+1) {
    for ( n=0; n<width_padded/block_size; n=n+1) {
      clear_matrix(block_img, block_size, block_size);
      for ( i=m*block_size; i<(m+1)*block_size; i++) {
	for ( j=n*block_size; j<(n+1)*block_size; j++) {
	  block_img[i-m*block_size][j-n*block_size] = img3[i][j];
	}
      }
      haar_wave_recon(block_img, lprfilter, hprfilter,FILTERLEN_L, 
		 FILTERLEN_H, block_size, block_size, num_levels);
      for ( i=m*block_size; i<(m+1)*block_size; i++) {
	for ( j=n*block_size; j<(n+1)*block_size; j++) {
	  img3[i][j] = block_img[i-m*block_size][j-n*block_size];
	}
      }
    }
  }
  multifree(block_img, 2);
}



void free_matrix(float ** matrix, int height)
{
  int i;
  for ( i=0; i<height; i++) {
    free(matrix[i]);
  }
}

int RoundTZero(float num)
{
  int intnum;
  if (num<0) {
    intnum = ceil(num);
    return intnum;
  }
  else {
    intnum = floor(num);
    return intnum;
  }
}



void OldApplyScaling(float **matrix, float *scaling_factors, 
                  int num_levels, int height, int width)
{
  int rlb, rub, clb, cub, level;
  int i, j;
  rlb = 0; rub = (height>>num_levels); 
  clb = 0; cub = (width>>num_levels);
  for ( i=rlb; i<rub; i++) {
    for ( j=clb; j<cub; j++) {
	matrix[i][j] = matrix[i][j]/scaling_factors[0];
    }
  }
  for ( level=1; level<=num_levels; level++) {
    rlb = 0; rub = (height>>(num_levels-level+1)); 
    clb = (width>>(num_levels-level+1)); cub = (width>>(num_levels-level));
    for ( i=rlb; i<rub; i++) {
      for ( j=clb; j<cub; j++) {
          matrix[i][j] = matrix[i][j]/scaling_factors[3*level-2];
      }
    }
    rlb = (height>>(num_levels-level+1)); rub = (height>>(num_levels-level));
    clb = 0; cub = (width>>(num_levels-level+1));
    for ( i=rlb; i<rub; i++) {
      for ( j=clb; j<cub; j++) {
         matrix[i][j] = matrix[i][j]/scaling_factors[3*level-1];
      }
    }
    rlb = (height>>(num_levels-level+1)); rub = (height>>(num_levels-level));
    clb = (width>>(num_levels-level+1)); cub = (width>>(num_levels-level));
    for ( i=rlb; i<rub; i++) {
      for ( j=clb; j<cub; j++) {
          matrix[i][j] = matrix[i][j]/scaling_factors[3*level];
      }
    }
  }

}





/***********************************************/
/*********** This is primary function *************/
/***********************************************/

void SpaceVaryingTransform(float **input_img, int height, int width, LIST **listmatrix, 
                           int height_padded, int width_padded, int block_size, int block_num_levels,   
                           float *scaling_factors, int num_levels, float **output_img, int num_of_stages)
{
  int i, j;
  int height_conved, width_conved;
  float **padded_img;
  clock_t time_1, time_2, time_3, time_4; 
  double cputime;
  int **band_size;
  int **band_pos;

  float lpfilter[2] = {1/sqrt(2), 1/sqrt(2)};
  float hpfilter[2] = {1/sqrt(2), -1/sqrt(2)};
  float lprfilter[2] = {1/sqrt(2), 1/sqrt(2)};
  float hprfilter[2] = {-1/sqrt(2), 1/sqrt(2)};
  

  band_size = multialloc(sizeof(int), 2, num_levels+1, 2);
  band_pos = multialloc(sizeof(int), 2, num_levels+1, 2);

  band_size[0][0] = height;
  band_size[0][1] = width;
  for ( i=1; i<=num_levels; i=i+1 ) {
    band_size[i][0] = (int)floor((band_size[i-1][0]+1)/2);
    band_size[i][1] = (int)floor((band_size[i-1][1]+1)/2);
  }
  band_pos[num_levels][0] = band_size[num_levels][0];
  band_pos[num_levels][1] = band_size[num_levels][1];
  for ( i=num_levels-1; i>=0; i=i-1 ) {
    band_pos[i][0] = band_pos[i+1][0]+band_size[i+1][0];
    band_pos[i][1] = band_pos[i+1][1]+band_size[i+1][1];
  }
  height_conved = band_pos[0][0];
  width_conved = band_pos[0][1];


  time_1 = clock();
  padded_img = multialloc(sizeof(float), 2, height_padded, width_padded);
  DecodeMultiply(listmatrix, height_conved, width_conved, height_padded, width_padded, 
                 band_size, band_pos, input_img, padded_img, num_levels, scaling_factors);
  time_3 = clock();
  
  if (num_of_stages==2) {
    BlockInverseTransform(padded_img, height_padded, width_padded, block_size, block_num_levels, 
			  lprfilter, hprfilter);
  }
  
  time_4 = clock();
  cputime = (double) (time_4-time_3)/(double)CLOCKS_PER_SEC;
  printf("cpu time for inverse wavelet transform = %f\n", cputime);
  time_2 = clock();
  cputime = (double) (time_2-time_1)/(double)CLOCKS_PER_SEC;
  printf("CPU time for efficient computation is %e\n", cputime);  

  for (i=0; i<height; i++) {
    for (j=0; j<width; j++) {
      output_img[i][j] = padded_img[i][j];
    }
  }

  multifree(padded_img, 2);

}






