#include "cimage.h"

void initImage(cimage* imp, int height, int width)
{
    imp->height = height;
	imp->width = width;
	imp->ptr = (float*) malloc(height*width*sizeof(float));
    if (imp->ptr == NULL) {
        printf("not enough memory\n");
        exit(-1);
	}
}

void destroyImage(cimage* imp)
{
    free(imp->ptr);
}

void initUcharImage(cimage_uchar* imp, int height, int width)
{
    imp->height = height;
    imp->width = width;
    imp->ptr = (unsigned char*) malloc(height*width*sizeof(unsigned char));
    if (imp->ptr==NULL) {
        printf("not enough memory\n");
        exit(-1);
	}
}

void destroyUcharImage(cimage_uchar* imp)
{
    free(imp->ptr);
}

void initIntImage(cimage_int* imp, int height, int width)
{
    imp->height = height;
    imp->width = width;
    imp->ptr = (int*) malloc(height*width*sizeof(int));
    if (imp->ptr==NULL) {
        printf("not enough memory\n");
        exit(-1);
	}
}

void destroyIntImage(cimage_int* imp)
{
    free(imp->ptr);
}

void initLongImage(cimage_long* imp, int height, int width)
{
    imp->height = height;
    imp->width = width;
    imp->ptr = (long*) malloc(height*width*sizeof(long));
    if (imp->ptr==NULL) {
        printf("not enough memory\n");
        exit(-1);
	}
}

void destroyLongImage(cimage_long* imp)
{
    free(imp->ptr);
}

void createFullColorImg(cimage_fullcolor* imp, int height, int width)
{
    initUcharImage(&(imp->im_ch1), height, width);
    initUcharImage(&(imp->im_ch2), height, width);
    initUcharImage(&(imp->im_ch3), height, width);
}

void destroyFullColorImg(cimage_fullcolor* imp)
{
    free(imp->im_ch1.ptr);
    free(imp->im_ch2.ptr);
    free(imp->im_ch3.ptr);
}

/* Fast methods to convert to gray */
/* 1. gray = (30r + 59g + 11b)/100 */
/* 2. gray = (11r + 16g + 5b)/32 */
/* 3. gray = (3r + 6g + 1b)/10 */
void rgb2Gray(cimage_fullcolor* imp, cimage_uchar* imgray)
{
    int i, len;
    unsigned char *puchar_r, *puchar_g, *puchar_b, *pgray;
    const int rgb2graycoef[3] = {3, 6, 1};
    const int norm = 10;

    if (imgray->height!=imp->im_ch1.height || imgray->width!=imp->im_ch1.width) {
        printf("image dimension mismatch in rgb2Gray\n");
	}

    pgray = imgray->ptr;
    puchar_r = imp->im_ch1.ptr;
    puchar_g = imp->im_ch2.ptr;
    puchar_b = imp->im_ch3.ptr;
    len = imgray->height*imgray->width;
    i=0;

    while (i<len) {
        *(pgray++) = (unsigned char)(((int)(*(puchar_r++))*rgb2graycoef[0] + (int)(*(puchar_g++))*rgb2graycoef[1] + (int)(*(puchar_b++))*rgb2graycoef[2])/norm);
        i++;
	}
}

/* The function is only for debugging, and should be deleted eventually */
/*
void writeImTest(cimage *imp, char* filename)
{
    int height = imp->height;
    int width = imp->width;
    float* p = imp->ptr;
    int x, y;

    FILE* fp = fopen(filename, "w");
    for ( y=0; y<height; y++ ) {
        for ( x=0; x<width; x++ ) {
            fprintf(fp, "%7.4f\t", *(p++));
		}
        fprintf(fp, "\n");
	}
    fclose(fp);
}
*/
