#include <math.h>
#include <string.h>
#include "list.h"

void AddItem (LIST* linklistPTR, unsigned short int pixel_row, 
	      unsigned short int pixel_col, float value)
{
  NODE* ptr;
  ptr = malloc(sizeof(NODE));
  ptr->pixel_row = pixel_row;
  ptr->pixel_col = pixel_col;
  ptr->pixel_val = value;
  ptr->link = NULL;

    if (linklistPTR->head != NULL) {
      linklistPTR->tail->link = ptr;
      linklistPTR->tail = ptr;
    }
    else {
      linklistPTR->head = ptr;
      linklistPTR->tail = ptr;
    }
    linklistPTR->num_pix = linklistPTR->num_pix+1;
}

void ClearQueue (LIST* linklistPTR) 
{
  NODE* ptr;
    while (linklistPTR->head != NULL) {
      ptr = linklistPTR->head->link;
      free(linklistPTR->head);
      linklistPTR->head = ptr;
    }
}

void FurtherQuantizeListmatrix(LIST **fine_listmatrix, LIST **coarse_listmatrix, 
		               int height, int width, float delta)
{
  int i, j;
  NODE *pnode;

  for (i=0; i<height; i++) {
    for (j=0; j<width; j++) {
      pnode = fine_listmatrix[i][j].head;
      while (pnode!=NULL) {
	if (fabs(pnode->pixel_val)>delta) {
          AddItem(&coarse_listmatrix[i][j], pnode->pixel_row, pnode->pixel_col, pnode->pixel_val);
	}
	pnode = pnode->link;
      }
    }
  }
}

/***** record listmatrix *****/
void RecordListmatrix(FILE *fp, LIST ** listmatrix, 
		      int height, int width)
{
  int i, j;
  NODE * pnode;
  
  for ( i=0; i<height; i=i+1) {
    for ( j=0; j<width; j=j+1) {
	fprintf(fp, "%d\n", listmatrix[i][j].num_pix);
        for ( pnode = listmatrix[i][j].head; pnode!=NULL; pnode = pnode->link) {
	  fprintf(fp, "%d\n%d\n", pnode->pixel_row, pnode->pixel_col);
	  fprintf(fp, "%f\n", pnode->pixel_val);
        }
    }
  }
}

void ReadListmatrix(FILE *fp, LIST ** listmatrix, int height, int width)
{
  int i, j, n;
  int pix_row, pix_col, num_pix;
  float pix_val;

  for ( i=0; i<height; i=i+1) {
    for ( j=0; j<width; j=j+1) {
      fscanf(fp, "%d", &num_pix);
      for ( n=0; n<num_pix; n=n+1) {
	fscanf(fp, "%d", &pix_row);
	fscanf(fp, "%d", &pix_col);
	fscanf(fp, "%f", &pix_val);
	AddItem(&listmatrix[i][j], pix_row, pix_col, pix_val);
      }
    }
  }
}

void WriteListmatrixHeader(FILE *fp, struct parameter para, float delta, int num_levels, 
                           int block_num_levels, int height, int width, int block_size)
{
  fprintf(fp, "HEADER\n");
  fprintf(fp, "%e %e %f %f %f\n", para.a, para.b, para.f, para.g, delta);
  fprintf(fp, "%d %d\n", num_levels, block_num_levels);
  fprintf(fp, "%d %d %d\n", height, width, block_size);
  fprintf(fp, "BODY\n");
}


int ReadNCheckListmatrixHeader(FILE *fp, struct parameter para, float delta, int num_levels, 
                                int block_num_levels, int height, int width, int block_size)
{
  char txt[10];
  float a, b, f, g, r_delta;
  int r_num_levels, r_block_num_levels, r_height, r_width, r_block_size;

  fscanf(fp, "%s", txt);
  if (strncmp(txt, "HEADER", 6)!=0) {
    return 0;
  }
  fscanf(fp, "%e", &a);
  if (fabs(a-para.a)>fabs(para.a)*1e-2) {
    return 0;
  }
  fscanf(fp, "%e", &b);
  if (fabs(b-para.b)>fabs(para.b)*1e-2) {
    return 0;
  }
  fscanf(fp, "%f", &f);
  if (fabs(f-para.f)>fabs(para.f)*1e-3) {
    return 0;
  }
  fscanf(fp, "%f", &g);
  if (fabs(g-para.g)>fabs(para.g)*1e-3) {
    return 0;
  }
  fscanf(fp, "%f", &r_delta);
  if (fabs(r_delta-delta)>fabs(delta)*1e-3) {
    return 0;
  }
  fscanf(fp, "%d", &r_num_levels);
  if (r_num_levels!=num_levels) {
    return 0;
  }
  fscanf(fp, "%d", &r_block_num_levels);
  if (r_block_num_levels!=block_num_levels) {
    return 0;
  }
  fscanf(fp, "%d", &r_height);
  if (r_height!=height) {
    return 0;
  }
  fscanf(fp, "%d", &r_width);
  if (r_width!=width) {
    return 0;
  }
  fscanf(fp, "%d", &r_block_size);
  if (r_block_size!=block_size) {
    return 0;
  }
  fscanf(fp, "%s", txt);
  if (strncmp(txt, "BODY", 4)!=0) {
    return 0;
  }
  return 1;
}

int ReadListmatrixHeader(FILE *fp, struct parameter *para, float *r_delta, int *r_num_levels, 
                                int *r_block_num_levels, int *r_height, int *r_width, int *r_block_size)
{
  char txt[10];

  fscanf(fp, "%s", txt);
  if (strncmp(txt, "HEADER", 6)!=0) {
    return 0;
  }
  fscanf(fp, "%e", &(para->a));
  fscanf(fp, "%e", &(para->b));
  fscanf(fp, "%f", &(para->f));
  fscanf(fp, "%f", &(para->g));
  fscanf(fp, "%f", r_delta);
  fscanf(fp, "%d", r_num_levels);
  fscanf(fp, "%d", r_block_num_levels);
  fscanf(fp, "%d", r_height);
  fscanf(fp, "%d", r_width);
  fscanf(fp, "%d", r_block_size);
  fscanf(fp, "%s", txt);
  if (strncmp(txt, "BODY", 4)!=0) {
    return 0;
  }
  return 1;
}


