#include "psf.h"


float computepsf(PARAMETER para, struct power_LUT *powertable, 
		 float x0, float y0, float x, float y, float pixel_size)
{
  float r2, term1, term2, value, norm_constant;
  float temp[1];
  float maxval = (float)(powertable->samples-1)/powertable->rate;
  x = x*pixel_size;
  y = y*pixel_size;
  x0 = x0*pixel_size;
  y0 = y0*pixel_size;

  r2 = x0*x0+y0*y0;
 
  norm_constant = 3.1415926*(para.g+para.a*r2)*(para.g+para.b*r2)/(para.f-1);
  norm_constant = norm_constant/pixel_size/pixel_size;
 
  if (r2==0) {
    temp[0] = 1+(x*x+y*y)/(para.g*para.g);
    if (temp[0]>maxval) {
      value = 0;
    }
    else {
      power(temp, 1, powertable);
      value = 1/temp[0]/norm_constant;
    }
  }
  else {
    term1 = (x*x0+y*y0-r2)/(para.g+para.a*r2);
    term2 = (-x*y0+y*x0)/(para.g+para.b*r2);
    temp[0] = 1+(term1*term1+term2*term2)/r2;
    if (temp[0]>maxval) {
      value = 0;
    }
    else {
      power(temp, 1, powertable);
      value = 1/temp[0]/norm_constant;
    }
  }
  return value;

}

/* Rigorous convolution using traditional strategy */
void RigoConv(struct parameter para, struct power_LUT * powertable,  int height, int width, 
              float **img1, float *img2, int *imgentry, int imgnonzero, float pixel_size)
{
  int i, j, n;
  float x0, y0, x, y, x_org, y_org;
  float psf; 
  x_org = (float)(width-1)/2;
  y_org = (float)(height-1)/2;

  for (n=0; n<imgnonzero; n++) {
	img2[n] = 0;
  }
  
  for ( i=0; i<height; i++) {
    for ( j=0; j<width; j++) {
      x0 = (float)j-x_org;
      y0 = (float)i-y_org;
      for ( n=0; n<imgnonzero; n=n+1 ) {
	x = (float)imgentry[2*n+1]-x_org;
	y = (float)imgentry[2*n]-y_org;
	psf = computepsf(para, powertable, x0, y0, x, y, pixel_size);
	img2[n] += psf*img1[i][j];
      }

    }
  }

}
