#include "sparseWT.h"



void showvector(float *vec, int length)
{
  int i;
  for ( i=0; i<length; i=i+1) {
    printf("%4.2f ", vec[i]);
  }
  printf("\n");
}

void showmatrix(float **matrix, int height, int width)
{
  int i;
  for ( i=0; i<height; i=i+1) {
    showvector(matrix[i], width);
  }
}

void downsample(float *convdata, int width, int rownum, LINKLIST * sparse_col_matrix, int start)
{
  int i, col;
  for ( i=start; i<width; i=i+2) {
    if (convdata[i]!=0) {
      col = (i-start)/2;
      addentry(&sparse_col_matrix[col], rownum, convdata[i]);
    }
  }
}

void show_sparse_matrix(LINKLIST * matrix, int dimension)
{
  int i;
  ENTRY * ptr;
  for ( i=0; i<dimension; i=i+1) {
    ptr = matrix[i].head;
    while (ptr!=NULL) {
      printf("position %d, value %4.2f\n", ptr->pixel_pos, ptr->pixel_val);
      ptr = ptr->link;
    }
    printf("********************\n");
  }
}

void initial_sparse_matrix(LINKLIST *sparse_matrix, int dim)
{
  int i;
  for ( i=0; i<dim; i=i+1 ) {
    sparse_matrix[i].head = NULL;
    sparse_matrix[i].tail = NULL;
    sparse_matrix[i].num_pix = 0;
  }
}

void clear_sparse_matrix(LINKLIST *sparse_matrix, int dim)
{
  int i;
  for ( i=0; i<dim; i=i+1 ) {
    clearlinklist(&sparse_matrix[i]);
  }
}

void clear_vector(float *datavector, int length)
{
  int i;
  for ( i=0; i<length; i=i+1 ) {
    datavector[i] = 0;
  }
}

void clear_matrix(float **matrix, int height, int width)
{
  int i, j;
  for ( i=0; i<height; i=i+1 ) {
    for ( j=0; j<width; j=j+1 ) {
      matrix[i][j] = 0;
    }
  }
}

void copy_upsamp_col(float *datavector, float **matrix, int col, 
		     int row_start, int row_end, int start)
{
  int i;
  for ( i=0; i<row_end-row_start+1; i=i+1 ) {
    datavector[i*2+start] = matrix[i+row_start][col];
  }
}

void copy_upsamp_row(float *datavector, float **matrix, int row,
		     int col_start, int col_end, int start)
{
  int i;
  for ( i=0; i<col_end-col_start+1; i=i+1 ) {
    datavector[i*2+start] = matrix[row][i+col_start];
  }
}

void copy_sparse_matrix(LINKLIST *source_matrix, LINKLIST *dest_matrix, int length)
{
  int i;
  ENTRY * ptr;
  for ( i=0; i<length; i=i+1 ) {
    ptr = source_matrix[i].head;
    while (ptr!=NULL) {
      addentry(&dest_matrix[i], ptr->pixel_pos, ptr->pixel_val);
      ptr = ptr->link;
    }
  }
}

void sym_extend(LINKLIST *plist, int filterlen, int width)
{
  ENTRY * p;
  int len;
  if (filterlen%2 == 0) {
    len = filterlen/2;
    p = plist->head;
    while ( p!=NULL ) {
      if (p->pixel_pos<0 || p->pixel_pos>=width) {
        break;
      }
      if (p->pixel_pos<len) {
        addentry(plist, -p->pixel_pos-1, p->pixel_val);
      }
      if (p->pixel_pos >= width-len) {
        addentry(plist, width - p->pixel_pos + width-1, p->pixel_val);
      }
      p = p->link;
    }
  }
  else {
    len = (filterlen-1)/2;
    p = plist->head;
    while ( p!=NULL ) {
      /* printf("at pixel position %d\n", p->pixel_pos); */
      if (p->pixel_pos<0 || p->pixel_pos>=width) {
        break;
      }
      if ( (p->pixel_pos<=len) && (p->pixel_pos>0) ) {
        addentry(plist, -p->pixel_pos, p->pixel_val);
      }
      if ( (p->pixel_pos >= width-len-1) && (p->pixel_pos<width-1) ) {
        addentry(plist, width - p->pixel_pos + width-2, p->pixel_val);
      }
      p = p->link;
    }
  }
  
}

int determine_non0(LINKLIST* data, int filterlen, int start, int *position, int width)
{
  int i, origin;
  int num_non0;
  ENTRY *p;
  num_non0 = 0;
  p = data->head;
  if ( filterlen%2 == 0) {
    origin = filterlen/2-1;
  }
  else {
    origin = (filterlen-1)/2;
  }
  while (p!=NULL) {
    for ( i=0; i<filterlen; i=i+1 ) {
      if ( (p->pixel_pos+i-origin>=0) && (p->pixel_pos+i-origin<width) 
	   && ((p->pixel_pos+i-origin-start)%2==0) ) {
	if ( (num_non0==0) || 
	     ((num_non0>0)&&(p->pixel_pos+i-origin>position[num_non0-1])) ) {
	  position[num_non0] = p->pixel_pos+i-origin;
	  num_non0 = num_non0 + 1;
	}
      }
    }
    p = p->link;
  }
  return num_non0;
}

int find_position(int *position, int length, int index)
{
  int i;
  for ( i=0; i<length; i=i+1) {
    if ( position[i]==index ) {
      return i;
    }
  }
  return -1;
}


int conv_downsample(float *non0_vals, LINKLIST *plist, float *filter, 
		    int filterlen, int width, int start, int *position)
{
  int i, origin, num_non0, index;
  ENTRY *p;

  p = plist->head;

  num_non0 = determine_non0(plist, filterlen, start, position, width);
  if ( num_non0 == 0 ) {
    return 0;
  }
  
  for ( i=0; i<num_non0; i=i+1 ) {
    non0_vals[i] = 0;
  }
  if ( filterlen%2 == 0) {
    origin = filterlen/2-1;
  }
  else {
    origin = (filterlen-1)/2;
  }
  while (p!=NULL) {
    for ( i=0; i<filterlen; i=i+1 ) {
      if ( (p->pixel_pos+i-origin>=0) && (p->pixel_pos+i-origin<width) ) {
	if ( (p->pixel_pos+i-origin-start)%2==0 ) {
	  index = find_position(position, num_non0, p->pixel_pos+i-origin);
	  if (index == -1) {
	    printf("cannot find nonzero position\n");
	    return -1;
	  }
	  if (index>=num_non0) {
	    printf("error finding index of non0 entry\n");
	    return -2;
	  }
	  non0_vals[index] = non0_vals[index] + p->pixel_val*filter[i];
	}
      }
    }
    p = p->link;
  }

  return num_non0;
}

void myconv(float *convdata, LINKLIST *plist, float *filter, int filterlen, int width)
{
  int i, origin;
  ENTRY* p;
  p = plist->head;
  if ( filterlen%2 == 0) {
    origin = filterlen/2-1;
  }
  else {
    origin = (filterlen-1)/2;
  }
  while (p!=NULL) {
    for ( i=0; i<filterlen; i=i+1 ) {
      if ( (p->pixel_pos+i-origin>=0) && (p->pixel_pos+i-origin<width) ) {
        convdata[i+p->pixel_pos-origin] = convdata[i+p->pixel_pos-origin] 
	  + p->pixel_val*filter[i];
      }
    }
    p = p->link;
  }  

}

/* convolution with zero extension */
void zeroconv(float *convdata, float *data, float *filter, int filterlen, int width)
{
  int i, j, origin, index;
  if (filterlen%2 == 0) {
    origin = filterlen/2-1;
  }
  else {
    origin = (filterlen-1)/2;
  }
  for ( j=0; j<width; j=j+1 ) {
    for ( i=0; i<filterlen; i=i+1 ) {
      index = j+i-origin;
      if ( (index>=0) && (index<width) ) {
        convdata[index] = convdata[index] + data[j]*filter[i];
      }
    }
  }
}



void regconv(float *convdata, float *data, float *filter, int filterlen, int width)
{
  int i, j, origin;

  if ( filterlen%2 == 0) {
    origin = filterlen/2-1;
  }
  else {
    origin = (filterlen-1)/2;
  }
  for ( j=0; j<width; j=j+1) {
    for ( i=0; i<filterlen; i=i+1 ) {
      if ( (j+i-origin>=0) && (j+i-origin<width) ) {
        convdata[j+i-origin] = convdata[j+i-origin] + data[j]*filter[i];
      }
    }
  }
  
  if ( filterlen%2 == 0 ) {
    for ( j=0; j<=origin; j=j+1 ) {
      for ( i=0; i<filterlen; i=i+1 ) {
	if (-j-1+i-origin>=0 && -j-1+i-origin<width ) {
	  convdata[-j-1+i-origin] = convdata[-j-1+i-origin] + data[j]*filter[i];
	}
      }
    }
    for ( j=width-origin; j<width; j=j+1 ) {
      for ( i=0; i<filterlen; i=i+1 ) {
	if (width-j-1+i-origin<0 && width-j+width-1+i-origin>=0) {
	  convdata[width-j+width-1+i-origin] = convdata[width-j+width-1+i-origin] 
	    + data[j]*filter[i];
	}
      }
    }
  }
  else {
    for ( j=1; j<=origin; j=j+1 ) {
      for ( i=0; i<filterlen; i=i+1 ) {
	if (-j+i-origin>=0 && -j+i-origin<width ) {
	  convdata[-j+i-origin] = convdata[-j+i-origin] + data[j]*filter[i];
	}
      }
    }
    for ( j=width-origin-1; j<width-1; j=j+1 ) {
      for ( i=0; i<filterlen; i=i+1 ) {
	if ( width-j+width-2+i-origin<width && width-j+width-2+i-origin>=0 ) {
	  convdata[width-j+width-2+i-origin] = convdata[width-j+width-2+i-origin] + data[j]*filter[i];
	}
      }
    }
  }

}



/* take sparse row matrix, output sparse column matrix, filter + downsample */
/* row is the first dimention, width is the other dimention */
/* ext is a flag indicating whether to extend it or not */
void symconv(float *filter, LINKLIST * sparse_row_matrix, int filterlen, 
                 int row, int width, LINKLIST * sparse_col_matrix, int ext, int start)
{
  int i, j;
  float *convdata;
  int *position;
  int num_non0;
  position = malloc(sizeof(int)*width);
  convdata = malloc(sizeof(float)*width/2);
  for ( i=0; i<row; i=i+1) {
    if (ext > 0) {
      sym_extend(&sparse_row_matrix[i], ext, width);
    }

    num_non0 = conv_downsample(convdata, &sparse_row_matrix[i], filter, 
			       filterlen, width, start, position);

    for ( j=0; j<num_non0; j=j+1 ) {
      if (convdata[j]!=0) {
	addentry(&sparse_col_matrix[(position[j]-start)/2], i, convdata[j]);
      }
    }
    
  }
  free(position);
  free(convdata);
}





/* lprfilter is low-pass reconstruction filter, hprfilter is high-pass 
   reconstruction filter */
void wave_recon(float **wave_coef, float *lprfilter, float *hprfilter, 
		int filterlen_l, int filterlen_h, int height, int width, int levels)
{
  int i, j, n;
  int subwidth, subheight;
  float *upsampled;
  float *upsampconved_l;
  float *upsampconved_h;
  float **l;
  float **h;
  for ( n=0; n<levels; n=n+1) {
    subheight = height>>(levels-n);
    subwidth = width>>(levels-n);
    upsampconved_l = malloc(sizeof(float)*subheight*2);
    upsampconved_h = malloc(sizeof(float)*subheight*2);
    upsampled = malloc(sizeof(float)*subheight*2);
    l = multialloc(sizeof(float), 2, subheight*2, subwidth);
    h = multialloc(sizeof(float), 2, subheight*2, subwidth);
    clear_vector(upsampconved_l, subheight*2);
    clear_vector(upsampconved_h, subheight*2);
    for ( j=0; j<subwidth; j=j+1 ) {
      clear_vector(upsampled, subheight*2);
      copy_upsamp_col(upsampled, wave_coef, j, 0, subheight-1, 0);
      regconv(upsampconved_l, upsampled, lprfilter, filterlen_h, subheight*2);
      clear_vector(upsampled, subheight*2);
      copy_upsamp_col(upsampled, wave_coef, j, subheight, 2*subheight-1, 1);
      regconv(upsampconved_l, upsampled, hprfilter, filterlen_l, subheight*2);
      for ( i=0; i<2*subheight; i=i+1) {
	l[i][j] = upsampconved_l[i];
      }
      clear_vector(upsampconved_l, subheight*2);
    }

    for ( j=subwidth; j<2*subwidth; j=j+1 ) {
      clear_vector(upsampled, subheight*2);
      copy_upsamp_col(upsampled, wave_coef, j, 0, subheight-1, 0);
      regconv(upsampconved_h, upsampled, lprfilter, filterlen_h, subheight*2);
      clear_vector(upsampled, subheight*2);
      copy_upsamp_col(upsampled, wave_coef, j, subheight, 2*subheight-1, 1);
      regconv(upsampconved_h, upsampled, hprfilter, filterlen_l, subheight*2);
      for ( i=0; i<2*subheight; i=i+1) {
	h[i][j-subwidth] = upsampconved_h[i];
      }
      clear_vector(upsampconved_h, subheight*2);
    }
    free(upsampled);
    free(upsampconved_l);
    free(upsampconved_h);

    upsampconved_l = malloc(sizeof(float)*subwidth*2);
    upsampconved_h = malloc(sizeof(float)*subwidth*2);
    upsampled = malloc(sizeof(float)*subwidth*2);
    clear_vector(upsampconved_l, subwidth*2);
    clear_vector(upsampconved_h, subwidth*2);
    for ( i=0; i<2*subheight; i=i+1 ) {
      clear_vector(upsampled, subwidth*2);
      copy_upsamp_row(upsampled, l, i, 0, subwidth-1, 0);
      regconv(upsampconved_l, upsampled, lprfilter, filterlen_h, subwidth*2);
      clear_vector(upsampled, subwidth*2);
      copy_upsamp_row(upsampled, h, i, 0, subwidth-1, 1);
      regconv(upsampconved_h, upsampled, hprfilter, filterlen_l, subwidth*2);
      for ( j=0; j<2*subwidth; j=j+1 ) {
	wave_coef[i][j] = upsampconved_l[j]+upsampconved_h[j];
      }
      clear_vector(upsampconved_l, subwidth*2);
      clear_vector(upsampconved_h, subwidth*2);
    }
    multifree(l, 2);
    multifree(h, 2);
    
    free(upsampled);
    free(upsampconved_l);
    free(upsampconved_h);
  }
}

void haar_wave_recon(float **wave_coef, float *lprfilter, float *hprfilter, 
		int filterlen_l, int filterlen_h, int height, int width, int levels)
{
  int i, j, n;
  int subwidth, subheight;
  float *upsampled;
  float *upsampconved_l;
  float *upsampconved_h;
  float **l;
  float **h;
  for ( n=0; n<levels; n=n+1) {
    subheight = height>>(levels-n);
    subwidth = width>>(levels-n);
    upsampconved_l = malloc(sizeof(float)*subheight*2);
    upsampconved_h = malloc(sizeof(float)*subheight*2);
    upsampled = malloc(sizeof(float)*subheight*2);
    l = multialloc(sizeof(float), 2, subheight*2, subwidth);
    h = multialloc(sizeof(float), 2, subheight*2, subwidth);
    clear_vector(upsampconved_l, subheight*2);
    clear_vector(upsampconved_h, subheight*2);
    for ( j=0; j<subwidth; j=j+1 ) {
      clear_vector(upsampled, subheight*2);
      copy_upsamp_col(upsampled, wave_coef, j, 0, subheight-1, 0);
      zeroconv(upsampconved_l, upsampled, lprfilter, filterlen_h, subheight*2);
      clear_vector(upsampled, subheight*2);
      copy_upsamp_col(upsampled, wave_coef, j, subheight, 2*subheight-1, 0);
      zeroconv(upsampconved_l, upsampled, hprfilter, filterlen_l, subheight*2);
      for ( i=0; i<2*subheight; i=i+1) {
	l[i][j] = upsampconved_l[i];
      }
      clear_vector(upsampconved_l, subheight*2);
    }

    for ( j=subwidth; j<2*subwidth; j=j+1 ) {
      clear_vector(upsampled, subheight*2);
      copy_upsamp_col(upsampled, wave_coef, j, 0, subheight-1, 0);
      zeroconv(upsampconved_h, upsampled, lprfilter, filterlen_h, subheight*2);
      clear_vector(upsampled, subheight*2);
      copy_upsamp_col(upsampled, wave_coef, j, subheight, 2*subheight-1, 0);
      zeroconv(upsampconved_h, upsampled, hprfilter, filterlen_l, subheight*2);
      for ( i=0; i<2*subheight; i=i+1) {
	h[i][j-subwidth] = upsampconved_h[i];
      }
      clear_vector(upsampconved_h, subheight*2);
    }
    free(upsampled);
    free(upsampconved_l);
    free(upsampconved_h);

    upsampconved_l = malloc(sizeof(float)*subwidth*2);
    upsampconved_h = malloc(sizeof(float)*subwidth*2);
    upsampled = malloc(sizeof(float)*subwidth*2);
    clear_vector(upsampconved_l, subwidth*2);
    clear_vector(upsampconved_h, subwidth*2);
    for ( i=0; i<2*subheight; i=i+1 ) {
      clear_vector(upsampled, subwidth*2);
      copy_upsamp_row(upsampled, l, i, 0, subwidth-1, 0);
      zeroconv(upsampconved_l, upsampled, lprfilter, filterlen_h, subwidth*2);
      clear_vector(upsampled, subwidth*2);
      copy_upsamp_row(upsampled, h, i, 0, subwidth-1, 0);
      zeroconv(upsampconved_h, upsampled, hprfilter, filterlen_l, subwidth*2);
      for ( j=0; j<2*subwidth; j=j+1 ) {
	wave_coef[i][j] = upsampconved_l[j]+upsampconved_h[j];
      }
      clear_vector(upsampconved_l, subwidth*2);
      clear_vector(upsampconved_h, subwidth*2);
    }
    multifree(l, 2);
    multifree(h, 2);
    
    free(upsampled);
    free(upsampconved_l);
    free(upsampconved_h);
  }
}


void sparse_decomp(float *lpfilter, float *hpfilter, LINKLIST *sparse_row_matrix, 
		   LINKLIST *wave_coef, int height, int width, int levels, 
		   int filterlen_l, int filterlen_h)
{
  int i, j;
  int subheight, subwidth, extlen;
  ENTRY *ptr;
  LINKLIST *sparse_col_matrix, *temp_matrix;
  LINKLIST *sparse_a, *sparse_h, *sparse_v, *sparse_d;
  temp_matrix = malloc(sizeof(LINKLIST)*height);
  initial_sparse_matrix(temp_matrix, height);
  copy_sparse_matrix(sparse_row_matrix, temp_matrix, height);
  if (filterlen_l>filterlen_h) {
    extlen = filterlen_l;
  }
  else {
    extlen = filterlen_h;
  }

  for ( i=0; i<levels; i=i+1 ) {
    subheight = (height>>i);
    subwidth = (width>>i);
    sparse_col_matrix = malloc(sizeof(LINKLIST)*subwidth/2);
    sparse_a = malloc(sizeof(LINKLIST)*subheight/2);
    sparse_h = malloc(sizeof(LINKLIST)*subheight/2);
    sparse_v = malloc(sizeof(LINKLIST)*subheight/2);
    sparse_d = malloc(sizeof(LINKLIST)*subheight/2);
    initial_sparse_matrix(sparse_col_matrix, subwidth/2);
    initial_sparse_matrix(sparse_a, subheight/2);
    initial_sparse_matrix(sparse_h, subheight/2);
    initial_sparse_matrix(sparse_v, subheight/2);
    initial_sparse_matrix(sparse_d, subheight/2);
    
    symconv(lpfilter, temp_matrix, filterlen_l, subheight, subwidth, sparse_col_matrix, extlen, 0);
    symconv(lpfilter, sparse_col_matrix, filterlen_l, subwidth/2, subheight, sparse_a, extlen, 0);
    symconv(hpfilter, sparse_col_matrix, filterlen_h, subwidth/2, subheight, sparse_v, 0, 1);
    clear_sparse_matrix(sparse_col_matrix, subwidth/2);
    initial_sparse_matrix(sparse_col_matrix, subwidth/2);
    symconv(hpfilter, temp_matrix, filterlen_h, subheight, subheight, sparse_col_matrix, 0, 1);
    symconv(lpfilter, sparse_col_matrix, filterlen_l, subwidth/2, subheight, sparse_h, extlen, 0);
    symconv(hpfilter, sparse_col_matrix, filterlen_h, subwidth/2, subheight, sparse_d, 0, 1);
    /*
    printf("show approximation coefficients\n");
    show_sparse_matrix(sparse_a, subheight/2);
    printf("show vertical entries\n");
    show_sparse_matrix(sparse_v, subheight/2);
    printf("show horizontal entries\n");
    show_sparse_matrix(sparse_h, subheight/2);
    printf("show diagnal entries\n");
    show_sparse_matrix(sparse_d, subheight/2);
    */

    clear_sparse_matrix(temp_matrix, subheight);
    free(temp_matrix);
    temp_matrix = malloc(sizeof(LINKLIST)*subheight/2);
    initial_sparse_matrix(temp_matrix, subheight/2);
    copy_sparse_matrix(sparse_a, temp_matrix, subheight/2);
    
    for ( j=0; j<subheight/2; j=j+1) {
      ptr = sparse_v[j].head;
      while(ptr!=NULL) {
        addentry(&wave_coef[j+subheight/2], ptr->pixel_pos, ptr->pixel_val);
        ptr = ptr->link;
      }
    }
    for ( j=0; j<subheight/2; j=j+1) {
      ptr = sparse_h[j].head;
      while (ptr!=NULL) {
        addentry(&wave_coef[j], ptr->pixel_pos+subwidth/2, ptr->pixel_val);
        ptr = ptr->link;
      }
    }
    for ( j=0; j<subheight/2; j=j+1) {
      ptr = sparse_d[j].head;
      while (ptr!=NULL) {
        addentry(&wave_coef[j+subheight/2], ptr->pixel_pos+subwidth/2, ptr->pixel_val); 
        ptr = ptr->link;
      }
    }
    
    clear_sparse_matrix(sparse_col_matrix, subwidth/2);
    clear_sparse_matrix(sparse_a, subheight/2);
    clear_sparse_matrix(sparse_h, subheight/2);
    clear_sparse_matrix(sparse_v, subheight/2);
    clear_sparse_matrix(sparse_d, subheight/2);
    free(sparse_col_matrix);
    free(sparse_a);
    free(sparse_h);
    free(sparse_v);
    free(sparse_d);
  }
  for ( j=0; j<subheight/2; j=j+1) {
      ptr = temp_matrix[j].head;
      while(ptr!=NULL) {
        addentry(&wave_coef[j], ptr->pixel_pos, ptr->pixel_val);
        ptr = ptr->link;
      }
  }
  clear_sparse_matrix(temp_matrix, subheight/2);
  free(temp_matrix);
}

void sparse_haar_decomp(float *lpfilter, float *hpfilter, LINKLIST *sparse_row_matrix, 
		   LINKLIST *wave_coef, int height, int width, int levels, 
		   int filterlen_l, int filterlen_h)
{
  int i, j;
  int subheight, subwidth, extlen;
  ENTRY *ptr;
  LINKLIST *sparse_col_matrix, *temp_matrix;
  LINKLIST *sparse_a, *sparse_h, *sparse_v, *sparse_d;
  temp_matrix = malloc(sizeof(LINKLIST)*height);
  initial_sparse_matrix(temp_matrix, height);
  copy_sparse_matrix(sparse_row_matrix, temp_matrix, height);
  if (filterlen_l>filterlen_h) {
    extlen = filterlen_l;
  }
  else {
    extlen = filterlen_h;
  }

  for ( i=0; i<levels; i=i+1 ) {
    subheight = (height>>i);
    subwidth = (width>>i);
    sparse_col_matrix = malloc(sizeof(LINKLIST)*subwidth/2);
    sparse_a = malloc(sizeof(LINKLIST)*subheight/2);
    sparse_h = malloc(sizeof(LINKLIST)*subheight/2);
    sparse_v = malloc(sizeof(LINKLIST)*subheight/2);
    sparse_d = malloc(sizeof(LINKLIST)*subheight/2);
    initial_sparse_matrix(sparse_col_matrix, subwidth/2);
    initial_sparse_matrix(sparse_a, subheight/2);
    initial_sparse_matrix(sparse_h, subheight/2);
    initial_sparse_matrix(sparse_v, subheight/2);
    initial_sparse_matrix(sparse_d, subheight/2);
    
    symconv(lpfilter, temp_matrix, filterlen_l, subheight, subwidth, sparse_col_matrix, 0, 1);
    symconv(lpfilter, sparse_col_matrix, filterlen_l, subwidth/2, subheight, sparse_a, 0, 1);
    symconv(hpfilter, sparse_col_matrix, filterlen_h, subwidth/2, subheight, sparse_v, 0, 1);
    clear_sparse_matrix(sparse_col_matrix, subwidth/2);
    initial_sparse_matrix(sparse_col_matrix, subwidth/2);
    symconv(hpfilter, temp_matrix, filterlen_h, subheight, subheight, sparse_col_matrix, 0, 1);
    symconv(lpfilter, sparse_col_matrix, filterlen_l, subwidth/2, subheight, sparse_h, 0, 1);
    symconv(hpfilter, sparse_col_matrix, filterlen_h, subwidth/2, subheight, sparse_d, 0, 1);
    /*
    printf("show approximation coefficients\n");
    show_sparse_matrix(sparse_a, subheight/2);
    printf("show vertical entries\n");
    show_sparse_matrix(sparse_v, subheight/2);
    printf("show horizontal entries\n");
    show_sparse_matrix(sparse_h, subheight/2);
    printf("show diagnal entries\n");
    show_sparse_matrix(sparse_d, subheight/2);
    */

    clear_sparse_matrix(temp_matrix, subheight);
    free(temp_matrix);
    temp_matrix = malloc(sizeof(LINKLIST)*subheight/2);
    initial_sparse_matrix(temp_matrix, subheight/2);
    copy_sparse_matrix(sparse_a, temp_matrix, subheight/2);
    
    for ( j=0; j<subheight/2; j=j+1) {
      ptr = sparse_v[j].head;
      while(ptr!=NULL) {
        addentry(&wave_coef[j+subheight/2], ptr->pixel_pos, ptr->pixel_val);
        ptr = ptr->link;
      }
    }
    for ( j=0; j<subheight/2; j=j+1) {
      ptr = sparse_h[j].head;
      while (ptr!=NULL) {
        addentry(&wave_coef[j], ptr->pixel_pos+subwidth/2, ptr->pixel_val);
        ptr = ptr->link;
      }
    }
    for ( j=0; j<subheight/2; j=j+1) {
      ptr = sparse_d[j].head;
      while (ptr!=NULL) {
        addentry(&wave_coef[j+subheight/2], ptr->pixel_pos+subwidth/2, ptr->pixel_val); 
        ptr = ptr->link;
      }
    }
    
    clear_sparse_matrix(sparse_col_matrix, subwidth/2);
    clear_sparse_matrix(sparse_a, subheight/2);
    clear_sparse_matrix(sparse_h, subheight/2);
    clear_sparse_matrix(sparse_v, subheight/2);
    clear_sparse_matrix(sparse_d, subheight/2);
    free(sparse_col_matrix);
    free(sparse_a);
    free(sparse_h);
    free(sparse_v);
    free(sparse_d);
  }
  for ( j=0; j<subheight/2; j=j+1) {
      ptr = temp_matrix[j].head;
      while(ptr!=NULL) {
        addentry(&wave_coef[j], ptr->pixel_pos, ptr->pixel_val);
        ptr = ptr->link;
      }
  }
  clear_sparse_matrix(temp_matrix, subheight/2);
  free(temp_matrix);
}


/*
int main(int argc, char *argv[])
{
  FILE *fp, *fq;
  LINKLIST *sparse_row_matrix;
 
  float **image;
  int i, j;
  float temp;
  float *lpfilter = malloc(sizeof(float)*FILTERLEN_L);
  float *hpfilter = malloc(sizeof(float)*FILTERLEN_H);
  float *lprfilter = malloc(sizeof(float)*FILTERLEN_H);
  float *hprfilter = malloc(sizeof(float)*FILTERLEN_L);
  image = multialloc(sizeof(float), 2, HEIGHT, WIDTH);
  sparse_row_matrix = malloc(sizeof(LINKLIST)*HEIGHT);

  initial_sparse_matrix(sparse_row_matrix, HEIGHT);

  if ( (fp = fopen(argv[1], "r")) ) {
    for ( i=0; i<HEIGHT; i=i+1 ) {
      for ( j=0; j<WIDTH; j=j+1) {
	fscanf(fp, "%f", &temp);
	if (temp!=0) {
	  addentry(&sparse_row_matrix[i], j, temp);
	}
      }
    }
    fclose(fp);
    printf("Done reading datafile %s\n", argv[1]);
  }
  else {
    printf("Cannot open datafile %s\n", argv[1]);
    exit(1);
  }

  if ( (fq = fopen("filter.dat", "r")) ) {
    for ( i=0; i<FILTERLEN_L; i=i+1 ) {
      fscanf(fq, "%f", &lpfilter[i]);
    }
    for ( i=0; i<FILTERLEN_H; i=i+1 ) {
      fscanf(fq, "%f", &hpfilter[i]);
    }
    for ( i=0; i<FILTERLEN_H; i=i+1 ) {
      fscanf(fq, "%f", &lprfilter[i]);
    }
    for ( i=0; i<FILTERLEN_L; i=i+1 ) {
      fscanf(fq, "%f", &hprfilter[i]);
    }
    fclose(fq);
    printf("Done reading filter file\n");
  }
  else {
    printf("Cannot open filter file\n");
    exit(1);
  }


  for ( i=0; i<HEIGHT; i=i+1) {
    for ( j=0; j<WIDTH; j=j+1) {
      image[i][j] = 0;
    }
  }


  sparse_decomp(lpfilter, hpfilter, sparse_row_matrix, image, HEIGHT, WIDTH, LEVELS, FILTERLEN_L, FILTERLEN_H);
  wave_recon(image, lprfilter, hprfilter, FILTERLEN_L, FILTERLEN_H, HEIGHT, WIDTH, LEVELS);

  if ( (fq = fopen("reconstructed.txt", "w")) ) {
    for ( i=0; i<HEIGHT; i=i+1 ) {
      for ( j=0; j<WIDTH; j=j+1 ) {
	fprintf(fq, "%12.4f  ", image[i][j]);
      }
      fprintf(fq, "\n");
    }
  }
  else {
    printf("cannot open file to write reconstructed data\n");
    exit(1);
  }


  clear_sparse_matrix(sparse_row_matrix, HEIGHT);
   
  free(sparse_row_matrix);
  free(lpfilter);
  free(hpfilter);
  free(lprfilter);
  free(hprfilter);
  multifree(image, 2);
  return 0;
}
*/
