/* test whether the code works for Haar wavelet instead of CDF5-3 wavelet for W */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <time.h>

#include "sparseWT.h"
#include "power.h"
#include "list.h"
#include "sparselink.h"

#define BLOCK_SIZE 64
#define BLOCK_LEVELS 4
#define FILTERLEN_L 2
#define FILTERLEN_H 2



int main(int argc, char *argv[])
{
  int i, j;
  float number;
  LINKLIST block_wavecoef[BLOCK_SIZE];
  LINKLIST sparse_column[BLOCK_SIZE];
  float **wave_coef;
  FILE *fp, *fq;
  ENTRY *fe;
  float lpfilter[2] = {1/sqrt(2), 1/sqrt(2)};
  float hpfilter[2] = {1/sqrt(2), -1/sqrt(2)};
  float lprfilter[2] = {1/sqrt(2), 1/sqrt(2)};
  float hprfilter[2] = {-1/sqrt(2), 1/sqrt(2)};


  initial_sparse_matrix(block_wavecoef, BLOCK_SIZE);
  initial_sparse_matrix(sparse_column, BLOCK_SIZE);
  wave_coef = multialloc(sizeof(float), 2, BLOCK_SIZE, BLOCK_SIZE);

  for ( i=0; i<BLOCK_SIZE; i++) {
    for ( j=0; j<BLOCK_SIZE; j++) {
      wave_coef[i][j] = 0;
    }
  }

  if ( (fp = fopen(argv[1], "r")) ) {
    for (i=0; i<BLOCK_SIZE; i++) {
      for (j=0; j<BLOCK_SIZE; j++) {
        fscanf(fp, "%f", &number);
        if (number != 0) {
          addentry(&sparse_column[i], j, number);
        }
      }
    }
    fclose(fp);
  }
  else {
    printf("cannot open data file\n");
    exit(1);
  }
  
  /* 
  printf("original matrix is as follows: \n");
  show_sparse_matrix(sparse_column, BLOCK_SIZE);
  */
  sparse_haar_decomp(lpfilter, hpfilter, sparse_column, block_wavecoef, BLOCK_SIZE, BLOCK_SIZE, 
                BLOCK_LEVELS, FILTERLEN_L, FILTERLEN_H);

  /*
  printf("\n transformed matrix is as follows: \n");
  show_sparse_matrix(block_wavecoef, BLOCK_SIZE);
  */

  for (i=0; i<BLOCK_SIZE; i++) {
    fe = block_wavecoef[i].head;
    while (fe!=NULL) {
      wave_coef[i][fe->pixel_pos] = fe->pixel_val;
      fe = fe->link;
    }
  }

  haar_wave_recon(wave_coef, lprfilter, hprfilter, FILTERLEN_L, FILTERLEN_H, BLOCK_SIZE, 
                  BLOCK_SIZE, BLOCK_LEVELS); 
  fq = fopen("recon.txt", "w");
  for (i=0; i<BLOCK_SIZE; i++) {
    for (j=0; j<BLOCK_SIZE; j++) {
      fprintf(fq, "%f\t", wave_coef[i][j]);
    }
    fprintf(fq, "\n");
  }
  fclose(fq);
  clear_sparse_matrix(sparse_column, BLOCK_SIZE);
  clear_sparse_matrix(block_wavecoef, BLOCK_SIZE);
  multifree(wave_coef, 2);
  return 0;
}
